#ifndef __PNGENCOD_H
#define __PNGENCOD_H
//
// Copyright (c) 1997,1998 Colosseum Builders, Inc.
// All rights reserved.
//
// Colosseum Builders, Inc. makes no warranty, expressed or implied
// with regards to this software. It is provided as is.
//
// See the README.TXT file that came with this software for restrictions
// on the use and redistribution of this file or send E-mail to
// info@colosseumbuilders.com
//

//
//  Title: PNG Encoder Class Definition
//
//  Author:  John M. Miano  miano@colosseumbuilders.com
//
//  Description:
//
//    This class implements PNG encoding.
//

#include "png.h"
#include "bitimage.h"
#include "acencod.h"

typedef double lenCodeLiteralBlockFilter[256];

class PngHuffmanEncoder ;

class PngEncoder : public BitmapImageEncoder
{
public:
  PngEncoder () ;
  PngEncoder (const PngEncoder &) ;
  virtual ~PngEncoder () ;

  PngEncoder &operator=(const PngEncoder &) ;

  void WriteImage (std::ostream &, BitmapImage &) ;

  // Property functions for predefined tEXt strings.
  void SetTitle (const std::string &) ;
  std::string GetTitle () const ;

  void SetAuthor (const std::string &) ;
  std::string GetAuthor () const ;

  void SetDescription (const std::string &) ;
  std::string GetDescription () const ;

  void SetCopyright (const std::string &) ;
  std::string GetCopyright () const ;

  void SetSoftware (const std::string &) ;
  std::string GetSoftware () const ;

  void SetDisclaimer (const std::string &) ;
  std::string GetDisclaimer () const ;

  void SetWarning (const std::string &) ;
  std::string GetWarning () const ;

  void SetSource (const std::string &) ;
  std::string GetSource () const ;

  void SetComment (const std::string &) ;
  std::string GetComment () const ;

  unsigned long GetBlockSize () const ;
  void SetBlockSize (unsigned long) ;

  enum CompressionLevel
  {
    FastestCompression = 0,
    FastCompression = 1,
    DefaultCompression = 2,
    MaximumCompression = 3,
  } ;

  void SetCompressionLevel (CompressionLevel) ;  //  
  CompressionLevel GetCompressionLevel () const ;
  bool compressionVertical, compressionGorizontal, compressionAverage, compressionDetect;
  void SetCompressionVertical();
  void SetCompressionGorizontal();
  void SetCompressionAverage();
  void SetCompressionDetect();

  void SetUseFilters (bool) ;
  bool GetUseFilters () const ;

  bool prevGreedyParsing; //     
  bool trebaBlockLongReplace; //    LZ-
  bool trebaOptimizeLZBlock;  //  LZ-
  bool trebaStatisticAnalizeLZBlock;  //   LZ-
  bool trebaPerexidCM; //     
  lenCodeLiteralBlockFilter * prognozLenCodeLiteralBF;
  UBYTE1 cm11, cm12, cm21, cm22; //     

protected:
  void Initialize () ;
  void DoCopy (const PngEncoder &) ;

private:
  unsigned long PngWindowSize; // -  PNG-

  struct HashEntry
  {
    unsigned long index;
    HashEntry *next ;
  };

  UBYTE1 *imageData;         //   
  typedef HashEntry *PHashEntry;
  HashEntry *hash_values;
  PHashEntry *hash_table;
  unsigned long *countHashValue;

  UBYTE1 countComponent;     //   
  UBYTE4 currentPozImage, countByteImage;
  UBYTE1 * tupBlockFilter;
  UBYTE4 * nextPozStartBlockFilter; //       
  UBYTE2 indexBlockFilter;

  void initCM();   //      
  void doneCM();   //    
  void writeCM();

  //      (MCU)
  void vstanovCMMCU(UBYTE2 shar, UBYTE2 indexBlockRow, UBYTE2 indexBlockCol, UBYTE2 startMCURow, UBYTE2 startMCUCol, UBYTE2 lenMCUShar);
  void writeCMMCU(UBYTE2 shar, UBYTE2 indexBlockRow, UBYTE2 indexBlockCol, UBYTE2 startMCURow, UBYTE2 startMCUCol, UBYTE2 lenMCUShar);
  void vstanovCMPixel(UBYTE1 nomer, UBYTE1 * pixel);

  void generateBlockLongReplace(); //    

  void compareStartPoslidovnist1(long adres1, long adres2, unsigned int &countRivni); //   
  void compareStartPoslidovnist2(long adres1, long adres2, unsigned int &countRivni); //   
  void compareStartPoslidovnist(long adres1, long adres2, unsigned int &countRivni, unsigned int bestLength); //   
  void compareStartPoslidovnistPixel(long adres1, long adres2, unsigned int &countRivni, unsigned int bestLength); //   
  void vcompareStartPoslidovnist1(long adres1, long adres2, unsigned int &countRivni); //   
  void vcompareStartPoslidovnist2(long adres1, long adres2, unsigned int &countRivni); //   
  void vcompareStartPoslidovnist(long adres1, long adres2, unsigned int &countRivni); //   
  void gcompareStartPoslidovnist1(long adres1, long adres2, unsigned int &countRivni); //   
  void gcompareStartPoslidovnist2(long adres1, long adres2, unsigned int &countRivni); //   
  void gcompareStartPoslidovnist(long adres1, long adres2, unsigned int &countRivni); //   
  void comparePoslidovnistImage(long adres1, long megaAdresProxid, long startOffset1, unsigned int countPixelLine,
                                unsigned int step, unsigned int startPixelX, unsigned int startPixelY, unsigned int &countRivni); //     

  void LongestMatch (unsigned int &length,
                     unsigned int &offset);
  void LongestFullMatch (unsigned int &length,
                         unsigned int &offset);
  void LongestMatchPixel(unsigned int &length,
                         unsigned int &offset);

  unsigned int HashValue (unsigned long) ;
  void FreePozHashEntry(unsigned long entry);
  void MoveHashEntry (unsigned long entry, unsigned int hashvalue) ;

  BitmapImage *current_image ;
  std::ostream *output_stream ;

  ACEncoder *distance_table ;
  ACEncoder *length_table ;
  bitBufferOutput *plusBitBuffer;
  PngHuffmanEncoder *length_length_table ;

  //  LZ-
  PngHuffmanEncoder *distance_literal_table;

  typedef void (PngEncoder::*LENGTHFUNCTION) (unsigned int index,
                                              unsigned int code,
                                              unsigned int length,
                                              unsigned int extra) ;
  void GatherLengthCounts (unsigned int,
                           unsigned int code,
                           unsigned int,
                           unsigned int) ;
  void OutputLengthCounts (unsigned int index,
                           unsigned int code,
                           unsigned int extra,
                           unsigned int value) ;

  void FindLengthCodes (LENGTHFUNCTION, UBYTE1 lengths [], unsigned int count) ;

  bool ProcessImageData () ;
  bool ProcessAnalizImageData () ;

  void plusPozImageComponent(unsigned int count);
  void ProcessImageDataComponent(unsigned int maxLen);
  void LongestMatchComponent (unsigned int &length,
                              unsigned int &offset,
                              unsigned int &plusByte,
                              unsigned int maxLenComponent);
  void VLongestMatchComponent(unsigned int &length,
                              unsigned int &offset,
                              unsigned int &plusByte,
                              unsigned int maxLenComponent);
  void GLongestMatchComponent(unsigned int &length,
                              unsigned int &offset,
                              unsigned int &plusByte,
                              unsigned int maxLenComponent);

  UBYTE1 *output_buffer ;
  UBYTE4 chunk_type ;
  UBYTE4 byte_position ;
  unsigned int bit_position ;
  void StartChunk (const char type [5]) ;
  void WriteCurrentChunk () ;
  void OutputBlock (const void *data, unsigned int length) ;
  void OutputByte (UBYTE1) ;
  void OutputDataBits (unsigned int data, unsigned int length) ;
  void OutputDataByte(UBYTE1 *buffer, UBYTE2 length);
  void FlushBitBuffer () ;
  void OutputAdler () ;

  void FreeBuffers () ;

  int row_width ; //  unsigned
  void WriteText (const std::string &keyword, const std::string &value) ;
  void WriteTextBlocks () ;

  void CallProgressFunction (unsigned int) ;

  std::string title_string ;
  std::string author_string ;
  std::string description_string ;
  std::string copyright_string ;
  std::string software_string ;
  std::string disclaimer_string ;
  std::string warning_string ;
  std::string source_string ;
  std::string comment_string ;

  enum { FilterBufferCount = 6, } ;
  UBYTE1 *filter_buffers [FilterBufferCount] ;
  unsigned int current_filter ;
  unsigned int filter_mask ;
  void FilterRow (unsigned int row) ;
  void FormFilterRow (unsigned int row, unsigned char filter) ;

  void DoWrite () ;

  void InitializeHashTable () ;
  void InitializeLZWindow () ;
  void OutputDeflateHeader (bool lastblock) ;
  void OutputZLibHeader () ;
  void WriteImageData () ;
  void prevGreedyParsingImageData(); //     

  CompressionLevel compression_level;
  unsigned int search_limit;
  UBYTE4 adler_value ;
  unsigned int image_row ;
  unsigned int image_col ;
  bool image_end ;

  UBYTE2 * block_buffer;
  bool trebaOptimizeBlock;     //   LZ- ,kjre
  PngHuffmanEncoder *length_literal_table ; //       <; >
  UBYTE1 * trebaZaminaBuffer;  // LZ-   
  UBYTE4 countAnalizZamina;    //     
  UBYTE4 * pozAnalizZamina;    //     
  UBYTE4 * pozImageBuffer;     //     
  unsigned int block_buffer_size ;
  unsigned int block_buffer_count ;

  void OutputBlockData () ;
  void SAOutputBlockData (bool lastblock) ;
} ;

inline void PngEncoder::SetTitle (const std::string &value)
{
  title_string = value ;
  return ;
}

inline std::string PngEncoder::GetTitle () const
{
  return title_string ;
}

inline void PngEncoder::SetAuthor (const std::string &value)
{
  author_string = value ;
  return ;
}

inline std::string PngEncoder::GetAuthor () const
{
  return author_string ;
}

inline void PngEncoder::SetDescription (const std::string &value)
{
  description_string = value ;
  return ;
}

inline std::string PngEncoder::GetDescription () const
{
  return description_string ;
}

inline void PngEncoder::SetCopyright (const std::string &value)
{
  copyright_string = value ;
  return ;
}

inline std::string PngEncoder::GetCopyright () const
{
  return copyright_string ;
}

inline void PngEncoder::SetSoftware (const std::string &value)
{
  software_string = value ;
  return ;
}

inline std::string PngEncoder::GetSoftware () const
{
  return software_string ;
}

inline void PngEncoder::SetDisclaimer (const std::string &value)
{
  disclaimer_string = value ;
  return ;
}

inline std::string PngEncoder::GetDisclaimer () const
{
  return disclaimer_string ;
}

inline void PngEncoder::SetWarning (const std::string &value)
{
  warning_string = value ;
  return ;
}

inline std::string PngEncoder::GetWarning () const
{
  return warning_string ;
}

inline void PngEncoder::SetSource (const std::string &value)
{
  source_string = value ;
  return ;
}

inline std::string PngEncoder::GetSource () const
{
  return source_string ;
}

inline void PngEncoder::SetComment (const std::string &value)
{
  comment_string = value ;
  return ;
}

inline std::string PngEncoder::GetComment () const
{
  return comment_string ;
}

inline unsigned long PngEncoder::GetBlockSize () const
{
  return block_buffer_size ;
}

inline void PngEncoder::SetBlockSize (unsigned long value)
{
  if (value < 500)
    block_buffer_size = 500 ;
  else
    block_buffer_size = value ;
  return ;
}

#endif


