#include <fstream>
#include <windows>
#include "pngencod.h"
#include "bmpdecod.h"

using namespace std ;
//extern unsigned long countFilter[FilterBufferCount];

void Progress (BitmapImageCoder &, void*, unsigned int pass, unsigned int passcount,
               unsigned int percent, bool &)
{
  unsigned int value = 100 * (pass - 1) / passcount + percent / passcount ;
  cout << value << "%                \r" << flush ;
}

void Usage (int argc, char *argv [])
{
  char *program ;
  if (argc != 0)
    program = argv [0] ;
  else
    program = "HBFENCOD" ;

  cerr << "Usage: " << program << " [-I -F -M -P -R -S -C] input.bmp output.hbf" << endl ;
  cerr << " -I - information of process" << endl ;
  cerr << " -F - fasted compression" << endl ;
  cerr << " -M - maximal compression" << endl ;
  cerr << " -P - not use predictory" << endl ;
  cerr << " -R - no analize replacement LZ-blocks" << endl; // LZ-,   
  cerr << " -S - statistics analize LZ-block" << endl;
  cerr << " -C - not to pass coloured model" << endl;

  exit (1) ;
  return ;
}

main (int argc, char *argv [])
{
  SYSTEMTIME startTime;
  GetLocalTime(&startTime);

  if (argc < 3)
    Usage (argc, argv) ;

  PngEncoder encoder ;

  for (unsigned int ii = 1 ; ii < argc - 2 ; ++ ii)
  {
    if (argv [ii][0] != '-')
      Usage (argc, argv) ;

    switch (argv [ii][1])
    {
    case 'P':
      encoder.SetUseFilters (false);
      break ;
    case 'F':
      encoder.SetCompressionLevel (PngEncoder::FastestCompression) ;
      break ;
    case 'M':
      encoder.SetCompressionLevel (PngEncoder::MaximumCompression) ;
      break ;
    case 'I':
      encoder.SetProgressFunction (Progress, NULL) ;
      break ;
    case 'R':
      encoder.trebaOptimizeLZBlock=false;
      break ;
    case 'S':
      encoder.trebaStatisticAnalizeLZBlock=true;
      break ;
    case 'C':
      encoder.trebaPerexidCM=false;
      break ;
    default:
      Usage (argc, argv) ;
    }
  }

  ifstream input (argv [argc-2], ios::binary) ;
  if (! input)
  {
    cerr << "Can't open input file " << argv [argc-2] << endl ;
    exit (1) ;
  }

  BmpDecoder decoder ;
  BitmapImage image ;
  try
  {
    decoder.ReadImage (input, image) ;
  }
  catch (exception &ee)
  {
    cerr << ee.what () << endl ;
	throw ;
  }
  char *name;
  int len=strlen(argv[argc-1]);
  name=new char[len+2];
  memcpy(name,argv[argc-1],len);
  name[len]='\0';
  ofstream output (name, ios::binary) ;
  if (! output)
  {
    cerr << "Can't open output file " << argv [argc-1] << endl ;
    exit (1) ;
  }
  delete[] name;
  name=NULL;
  if (image.BitCount() != 24)
   {cerr << "Image not is True Color (24 bits)" << endl ;
    exit (1) ;
   }

  try
  {
    encoder.WriteImage (output, image) ;
  }
  catch (exception &ee)
  {
    cerr << ee.what () << endl ;
  }
  SYSTEMTIME endTime;
  GetLocalTime(&endTime);
  long int interval= (endTime.wHour-startTime.wHour)*60*60*1000 +
                     (endTime.wMinute-startTime.wMinute)*60*1000 +
                     (endTime.wSecond-startTime.wSecond)*1000 +
                     (endTime.wMilliseconds-startTime.wMilliseconds);
  cout << endl << interval << " milliseconds " << endl ;
  return 0 ;
}
