//
// Copyright (c) 1997,1998 Colosseum Builders, Inc.
// All rights reserved.
//
// Colosseum Builders, Inc. makes no warranty, expressed or implied
// with regards to this software. It is provided as is.
//
// See the README.TXT file that came with this software for restrictions
// on the use and redistribution of this file or send E-mail to
// info@colosseumbuilders.com
//

//
//  Title:  PNG Encoder Class Implementation
//
//  Author:  John M. Miano miano@colosseumbuilders.com
//
#include <limits.h>
#include <windows.h>

#include "pngencod.h"
#include "pngpvt.h"
#include "pngchksm.h"
#include "pnghuffe.h"

using namespace std ;

// Hash Table size definitions.
const unsigned int HashBits = 5;
const unsigned char lenKeyHash = 3; //    -
const unsigned int HashTableSize = 1 << (3 * HashBits) ;
const int maxBlockBufferSize=0x10000;
const unsigned int minLenZaminaLiteral=18; //     ,   
unsigned long countZamina;
UBYTE1 *imageDataPredict; //    
UBYTE2 *countPixelLineBlockFilter; //      
UBYTE2 *stepBlockFilter; //    
UBYTE1 *proxidBlockFilter; //      

unsigned char codesDistance[32769];

// Size of the output buffer. This value defines the maximum size for
// IDAT block. This value must be larger than the size of any non-IDAT
// blocks used to write the image. Other than that the value can be
// selected arbitrarily.
const unsigned int OutputBufferSize = (1 << 16) ;

static inline Hash (UBYTE1 v1, UBYTE1 v2, UBYTE1 v3)
{
  const unsigned int mask = (1 << HashBits) - 1 ;
  unsigned int value = (v1 & mask )
                     | ((v2 & mask) << HashBits)
                     | ((v3 & mask) << (2 * HashBits)) ;
  return value ;

}

  // extras [n] is the number of extra bits for the code n.
  static const unsigned int extrasDistance [PngMaxDistanceCodes] =
    {
      0, 0,  0,  0,  1,  1,  2,  2,  3, 3,
      4, 4,  5,  5,  6,  6,  7,  7,  8, 8,
      9, 9, 10, 10, 11, 11, 12, 12, 13, 13,
    } ;

  // bases [n] is the smallest distance value for code n.
  static const unsigned int basesDistance [PngMaxDistanceCodes] =
    {
         1,    2,    3,    4,    5,    7,    9,    13,    17,    25,
        33,   49,   65,   97,  129,  193,  257,   385,   513,   769,
      1025, 1537, 2049, 3073, 4097, 6145, 8193, 12289, 16385, 24577,
    } ;

 void initCodesDistance()
  {// maxvalue [n] is the maximum distance value for the code n.
   static const unsigned int maxvalue [PngMaxDistanceCodes] =
    {
         1,     2,     3,     4,     6,     8,    12,    16,    24,    32,
        48,    64,    96,   128,   192,   256,   384,   512,   768,  1024,
      1536,  2048,  3072,  4096,  6144,  8192, 12288, 16384, 24576, 32768,
    } ;
   UBYTE2 distance=0;
   for (unsigned int code=0; code<PngMaxDistanceCodes; code++)
    while (distance<=maxvalue[code])
     codesDistance[distance++]=code;
  }

static void DistanceToCode (unsigned int distance, unsigned int &code,
                            unsigned int &extra, unsigned int &value)
{
  code=codesDistance[distance];
  extra = extrasDistance[code];
  value = distance - basesDistance[code];
  return ;
}

  // extras [n] is the number of extra bits for code n.
  static const UBYTE1 extrasLength [PngMaxLengthCodes - PngFirstLengthCode] =
    {
      0, 0, 0, 0, 0, 0, 0, 0,
      1, 1, 1, 1, 2, 2, 2, 2,
      3, 3, 3, 3, 4, 4, 4, 4,
      5, 5, 5, 5, 0, 0, 0,
    } ;
// codes [n] is the length for for length n - 3.
  static const UBYTE2 codesLength [PngLongestLength-2] =
    {
      257, 258, 259, 260, 261, 262, 263, 264,
      265, 265, 266, 266, 267, 267, 268, 268,
      269, 269, 269, 269, 270, 270, 270, 270,
      271, 271, 271, 271, 272, 272, 272, 272,
      273, 273, 273, 273, 273, 273, 273, 273,
      274, 274, 274, 274, 274, 274, 274, 274,
      275, 275, 275, 275, 275, 275, 275, 275,
      276, 276, 276, 276, 276, 276, 276, 276,
      277, 277, 277, 277, 277, 277, 277, 277,
      277, 277, 277, 277, 277, 277, 277, 277,
      278, 278, 278, 278, 278, 278, 278, 278,
      278, 278, 278, 278, 278, 278, 278, 278,
      279, 279, 279, 279, 279, 279, 279, 279,
      279, 279, 279, 279, 279, 279, 279, 279,
      280, 280, 280, 280, 280, 280, 280, 280,
      280, 280, 280, 280, 280, 280, 280, 280,
      281, 281, 281, 281, 281, 281, 281, 281,
      281, 281, 281, 281, 281, 281, 281, 281,
      281, 281, 281, 281, 281, 281, 281, 281,
      281, 281, 281, 281, 281, 281, 281, 281,
      282, 282, 282, 282, 282, 282, 282, 282,
      282, 282, 282, 282, 282, 282, 282, 282,
      282, 282, 282, 282, 282, 282, 282, 282,
      282, 282, 282, 282, 282, 282, 282, 282,
      283, 283, 283, 283, 283, 283, 283, 283,
      283, 283, 283, 283, 283, 283, 283, 283,
      283, 283, 283, 283, 283, 283, 283, 283,
      283, 283, 283, 283, 283, 283, 283, 283,
      284, 284, 284, 284, 284, 284, 284, 284,
      284, 284, 284, 284, 284, 284, 284, 284,
      284, 284, 284, 284, 284, 284, 284, 284,
      284, 284, 284, 284, 284, 284, 284, 285,
    } ;


inline static void LengthToCode (unsigned int length, unsigned int &code,
                                 unsigned int &extra, unsigned int &value)
{
  static const UBYTE2 basevalues [PngMaxLengthCodes - PngFirstLengthCode] =
    {
        3,   4,   5,   6,   7,   8,   9,  10,
       11,  13,  15,  17,  19,  23,  27,  31,
       35,  43,  51,  59,  67,  83,  99, 115,
      131, 163, 195, 227, 258,   0,   0,
    } ;

  code = codesLength [length - 3] ;
  extra = extrasLength [code - PngFirstLengthCode] ;
  value = length - basevalues [code - PngFirstLengthCode] ;
  return ;
}

//
//  Description:
//
//    Class Default Constructor
//
PngEncoder::PngEncoder ()
{
  Initialize () ;
  return ;
}

//
//  Description:
//
//    Class Copy Constructor
//
PngEncoder::PngEncoder (const PngEncoder &source)
{
  Initialize () ;
  DoCopy (source) ;
  return ;
}

//
//  Description:
//
//    Class Destructor
//
PngEncoder::~PngEncoder ()
{
  FreeBuffers () ;
  delete distance_table ;
  delete length_table ;
  delete plusBitBuffer;
  delete length_length_table ;
  return ;
}

//
//  Descriptor:
//
//    Assignment Operator
//
//  Parameters:
//    source: The object to copy
//
PngEncoder &PngEncoder::operator=(const PngEncoder &source)
{
  DoCopy (source) ;
  return *this ;
}

//
//  Description:
//
//    Common class initialization function. This function is
//    called from constructors to initialize the object.
//
void PngEncoder::Initialize ()
{
  compression_level=DefaultCompression;
  MakeCrcTable () ;
  software_string = "Colosseum Builders Image Library" ;
  trebaOptimizeLZBlock = true; //   LZ- 
  trebaStatisticAnalizeLZBlock = false; //    LZ-  
  trebaPerexidCM = true; //          
  current_image = NULL ;

  filter_mask = 0xFFFFFFFF ;
  for (unsigned int j = 0 ; j < FilterBufferCount ; ++ j)
    filter_buffers [j] = NULL ;

  distance_table = new ACEncoder ;
  length_table = new ACEncoder ;
  plusBitBuffer = new bitBufferOutput ;
  length_length_table = new PngHuffmanEncoder ;

  block_buffer = NULL ;
  trebaZaminaBuffer = NULL ;
  pozAnalizZamina = NULL;
  pozImageBuffer = NULL;

  block_buffer_size = maxBlockBufferSize;
  return ;
}

//
//  Description:
//
//    Class copy function. This function is called from the copy
//    constructor and assignment operator.
//
//  Parameters:
//    source:  The object to copy
//
void PngEncoder::DoCopy (const PngEncoder &source)
{
  title_string = source.title_string ;
  author_string = source.author_string ;
  description_string = source.description_string ;
  copyright_string = source.copyright_string ;
  software_string = source.software_string ;
  disclaimer_string = source.disclaimer_string ;
  warning_string = source.warning_string ;
  source_string = source.source_string ;
  comment_string = source.comment_string ;
  block_buffer_size = source.block_buffer_size ;
  BitmapImageEncoder::DoCopy (source) ;
  return ;
}

//
//  Description:
//
//    This function writes an image to a PNG stream.
//
//  Parameters:
//    strm:  The output stream
//    image:  The image to outpu
//
void PngEncoder::WriteImage (std::ostream &strm, BitmapImage &image)
{
  current_image = &image ;
  output_stream = &strm ;

  try
  {
    DoWrite () ;
  }
  catch (...)
  {
    FreeBuffers () ;
    throw ;
  }
  FreeBuffers () ;
  return ;
}

//
//  Description:
//
//    This function frees the buffers allowed during the encoding
//    process.
//
void PngEncoder::FreeBuffers ()
{
  current_image = NULL ;
  output_stream = NULL ;
  delete [] hash_values ; hash_values = NULL ;
  delete [] hash_table ; hash_table = NULL ;
  delete [] output_buffer ; output_buffer = NULL ;
  delete [] imageData; imageData = NULL;
  delete [] imageDataPredict; imageDataPredict=NULL;
  delete [] nextPozStartBlockFilter; nextPozStartBlockFilter=NULL;
  delete [] countPixelLineBlockFilter; countPixelLineBlockFilter=NULL;
  delete [] stepBlockFilter; stepBlockFilter=NULL;
  delete [] proxidBlockFilter; proxidBlockFilter=NULL;

  for (unsigned int j = 0 ; j < FilterBufferCount ; ++ j)
   {if (((1 << j) & filter_mask) != 0)
     delete [] filter_buffers[j];
    filter_buffers[j] = NULL;
   }
  return ;
}

//
//  Description:
//
//    This function calculates the hash value from a location
//    in the input stream. We create the hash value by
//    extracting a fixed number of the low order bits from each
//    of the next three data values.
//
//  Parameters:
//    index: The index into the lookahead buffer
//
//  Return Value:
//    The hash value
//
unsigned int PngEncoder::HashValue (unsigned long index)
{
  return Hash (imageData[index],
               imageData[index+1],
               imageData[index+2]) ;
}

//
//  Description:
//
//    This function finds the longest string in the LZ
//    window that matches the data in the lookahead buffer.
//
//    Since the hash changes can get very long under certain
//    circumstances we use the search_limit to set a limit
//    to how far we search in the tree.
//
//  Parameters:
//    bestlength: The longest match found (0=>No match)
//    bestoffset: The location where the best match was found
//

void PngEncoder::LongestMatch (unsigned int &bestlength,
                               unsigned int &bestoffset)
 {bestlength = 0;
  //return;
  unsigned long bestPoz, lenAnalizHash=lenKeyHash;
  //  -   
  unsigned int hashvalue = HashValue (currentPozImage);
  if (countHashValue[hashvalue] == 0)
   return;
  HashEntry *current;
  unsigned int countRivni, delta=0, countAnalizValue=0;
  unsigned long countZaluchokPoz;
  current=hash_table[hashvalue];
  countZaluchokPoz=countHashValue[hashvalue];
startAnalizPoz:
  while (countZaluchokPoz>0 && countAnalizValue++ < search_limit)
   {if (currentPozImage-(current->index-delta)>PngWindowSize)
     break; //    
    compareStartPoslidovnist(currentPozImage, current->index-delta, countRivni, bestlength);
    countZaluchokPoz--;
    if (((currentPozImage-current->index+delta)%3==0) && countRivni>bestlength)
     {if (bestlength>=3) //       
       if (countRivni-bestlength<=2) //    ,    
        if (currentPozImage-(current->index-delta)>256) // o    '  
         if (((countRivni-bestlength==1) && (currentPozImage-bestPoz<=256)) ||
             ((countRivni-bestlength==2) && (currentPozImage-bestPoz<=4)))
          if ((max(currentPozImage-bestPoz,(unsigned long)4) << (6*(countRivni-bestlength))) < currentPozImage-(current->index-delta))
           {current=current->next;
            continue;
           }
      bestlength=countRivni;
      bestPoz=current->index-delta;
      if (bestlength==PngLongestLength) //  
       break;
      if (countZaluchokPoz>0) //     
       {while (lenAnalizHash<=bestlength)
         {lenAnalizHash++;
          unsigned int newHashValue = HashValue (currentPozImage+lenAnalizHash-lenKeyHash);
          if (countZaluchokPoz>countHashValue[newHashValue])
           {countZaluchokPoz=countHashValue[newHashValue];
            current=hash_table[newHashValue];
            delta=lenAnalizHash-lenKeyHash;
            goto startAnalizPoz;
           }
         }
       }
     }
    current=current->next;
   }
  if (currentPozImage+bestlength > countByteImage)
   bestlength=countByteImage-currentPozImage;
  if (bestlength<3)
   {bestlength=0;
    return;
   }
  bestoffset=(currentPozImage-bestPoz)/3+4;
  if (indexBlockFilter>0) //  
   {UBYTE2 step=stepBlockFilter[indexBlockFilter];
    switch (proxidBlockFilter[indexBlockFilter])
     {case 1:
       comparePoslidovnistImage(currentPozImage, nextPozStartBlockFilter[indexBlockFilter], currentPozImage-nextPozStartBlockFilter[indexBlockFilter-1],
                                countPixelLineBlockFilter[indexBlockFilter], step,
                                0, 0, countRivni);
       if (countRivni>=bestlength)
        {bestlength=countRivni;
         bestoffset=1;
        }
       comparePoslidovnistImage(currentPozImage, nextPozStartBlockFilter[indexBlockFilter], currentPozImage-nextPozStartBlockFilter[indexBlockFilter-1],
                                countPixelLineBlockFilter[indexBlockFilter], step,
                                step, 0, countRivni);
       if (countRivni>bestlength ||
           countRivni==bestlength && bestoffset>4)
        {bestlength=countRivni;
         bestoffset=2;
        }
       comparePoslidovnistImage(currentPozImage, nextPozStartBlockFilter[indexBlockFilter], currentPozImage-nextPozStartBlockFilter[indexBlockFilter-1],
                                countPixelLineBlockFilter[indexBlockFilter], step,
                                step, step, countRivni);
       if (countRivni>bestlength ||
           countRivni==bestlength && bestoffset>4)
        {bestlength=countRivni;
         bestoffset=3;
        }
       comparePoslidovnistImage(currentPozImage, nextPozStartBlockFilter[indexBlockFilter], currentPozImage-nextPozStartBlockFilter[indexBlockFilter-1],
                                countPixelLineBlockFilter[indexBlockFilter], step,
                                0, step, countRivni);
       if (countRivni>bestlength ||
           countRivni==bestlength && bestoffset>4)
        {bestlength=countRivni;
         bestoffset=4;
        }
       break;
      case 2:
       comparePoslidovnistImage(currentPozImage, nextPozStartBlockFilter[indexBlockFilter], currentPozImage-nextPozStartBlockFilter[indexBlockFilter-1],
                                countPixelLineBlockFilter[indexBlockFilter], step,
                                -step/2, step/2, countRivni);
       if (countRivni>=bestlength)
        {bestlength=countRivni;
         bestoffset=1;
        }
       comparePoslidovnistImage(currentPozImage, nextPozStartBlockFilter[indexBlockFilter], currentPozImage-nextPozStartBlockFilter[indexBlockFilter-1],
                                countPixelLineBlockFilter[indexBlockFilter], step,
                                0, 0, countRivni);
       if (countRivni>bestlength ||
           countRivni==bestlength && bestoffset>4)
        {bestlength=countRivni;
         bestoffset=2;
        }
       comparePoslidovnistImage(currentPozImage, nextPozStartBlockFilter[indexBlockFilter], currentPozImage-nextPozStartBlockFilter[indexBlockFilter-1],
                                countPixelLineBlockFilter[indexBlockFilter], step,
                                step/2, step/2, countRivni);
       if (countRivni>bestlength ||
           countRivni==bestlength && bestoffset>4)
        {bestlength=countRivni;
         bestoffset=3;
        }
       comparePoslidovnistImage(currentPozImage, nextPozStartBlockFilter[indexBlockFilter], currentPozImage-nextPozStartBlockFilter[indexBlockFilter-1],
                                countPixelLineBlockFilter[indexBlockFilter], step,
                                0, step, countRivni);
       if (countRivni>bestlength ||
           countRivni==bestlength && bestoffset>4)
        {bestlength=countRivni;
         bestoffset=4;
        }
       break;
      case 3:
       comparePoslidovnistImage(currentPozImage, nextPozStartBlockFilter[indexBlockFilter], currentPozImage-nextPozStartBlockFilter[indexBlockFilter-1],
                                countPixelLineBlockFilter[indexBlockFilter], step,
                                0, 0, countRivni);
       if (countRivni>=bestlength)
        {bestlength=countRivni;
         bestoffset=1;
        }
       comparePoslidovnistImage(currentPozImage, nextPozStartBlockFilter[indexBlockFilter], currentPozImage-nextPozStartBlockFilter[indexBlockFilter-1],
                                countPixelLineBlockFilter[indexBlockFilter], step,
                                step/2, -step/2, countRivni);
       if (countRivni>bestlength ||
           countRivni==bestlength && bestoffset>4)
        {bestlength=countRivni;
         bestoffset=2;
        }
       comparePoslidovnistImage(currentPozImage, nextPozStartBlockFilter[indexBlockFilter], currentPozImage-nextPozStartBlockFilter[indexBlockFilter-1],
                                countPixelLineBlockFilter[indexBlockFilter], step,
                                step, 0, countRivni);
       if (countRivni>bestlength ||
           countRivni==bestlength && bestoffset>4)
        {bestlength=countRivni;
         bestoffset=3;
        }
       comparePoslidovnistImage(currentPozImage, nextPozStartBlockFilter[indexBlockFilter], currentPozImage-nextPozStartBlockFilter[indexBlockFilter-1],
                                countPixelLineBlockFilter[indexBlockFilter], step,
                                step/2, step/2, countRivni);
       if (countRivni>bestlength ||
           countRivni==bestlength && bestoffset>4)
        {bestlength=countRivni;
         bestoffset=4;
        }
       break;
      }
   }
  return ;
}

void PngEncoder::LongestMatchPixel(unsigned int &bestlength,
                                   unsigned int &bestoffset)
 {bestlength = 0;
  unsigned long bestPoz, lenAnalizHash=lenKeyHash;
  //  -   
  unsigned int hashvalue = HashValue (currentPozImage);
  if (countHashValue[hashvalue] == 0)
   return;
  HashEntry *current;
  unsigned int countRivni, delta=0, countAnalizValue=0;
  unsigned long countZaluchokPoz;
  current=hash_table[hashvalue];
  countZaluchokPoz=countHashValue[hashvalue];
startAnalizPoz:
  while (countZaluchokPoz>0 && countAnalizValue++ < search_limit)
   {if (currentPozImage-(current->index-delta)>PngWindowSize)
     break; //    
    compareStartPoslidovnist(currentPozImage, current->index-delta, countRivni, bestlength);
    countZaluchokPoz--;
    countRivni-=countRivni%3;
    if (countRivni>bestlength)
     {bestlength=countRivni;
      bestPoz=current->index-delta;
      if (bestlength==PngLongestLength) //  
       break;
      if (countZaluchokPoz>0) //     
       {while (lenAnalizHash<=bestlength)
         {lenAnalizHash+=3;
          unsigned int newHashValue = HashValue (currentPozImage+lenAnalizHash-lenKeyHash);
          if (countZaluchokPoz>countHashValue[newHashValue])
           {countZaluchokPoz=countHashValue[newHashValue];
            current=hash_table[newHashValue];
            delta=lenAnalizHash-lenKeyHash;
            goto startAnalizPoz;
           }
         }
       }
     }
    current=current->next;
   }
  if (currentPozImage+bestlength > countByteImage)
   bestlength=countByteImage-currentPozImage;
  if (bestlength<3)
   {bestlength=0;
    return;
   }
  bestoffset=currentPozImage-bestPoz;
  return ;
}

void PngEncoder::comparePoslidovnistImage(long adres1,
                                          long megaAdresProxid,
                                          long startOffset1,
                                          unsigned int countPixelLine,
                                          unsigned int step,
                                          unsigned int startPixelX,
                                          unsigned int startPixelY,
                                          unsigned int &countRivni)
 {countRivni=0;
  if (adres1<1)
   return;
  UBYTE1 delta;
  UBYTE2 rowPixelProxid, colPixelProxid;
  int row, col;
  for (countRivni=0 ; countRivni<PngLongestLength; countRivni++)
   {if (adres1+countRivni>=megaAdresProxid)
     return;
    delta=(startOffset1+countRivni) % 3;
    rowPixelProxid=((startOffset1+countRivni)/3)/countPixelLine;
    colPixelProxid=((startOffset1+countRivni)/3) % countPixelLine;
    row=rowPixelProxid*step+startPixelY;
    if (row<0 || row>=image_row)
     return;
    col=colPixelProxid*step+startPixelX;
    if (col<0 || col>=image_col)
     return;
    if (imageData[adres1+countRivni]!=(*current_image)[row][3*col+2-delta])
     return;
   }
  return;
 }

void PngEncoder::compareStartPoslidovnist(long adres1,
                                          long adres2,
                                          unsigned int &countRivni,
                                          unsigned int bestLength)
 {countRivni=0;
  if (adres1<1 || adres2<0)
   return;
  if (imageData[adres1+bestLength]!=imageData[adres2+bestLength]) //   
   return;
  for ( ; countRivni<PngLongestLength; countRivni++)
   if (imageData[adres1+countRivni]!=imageData[adres2+countRivni])
    return;
  return;
 }

//    
void PngEncoder::FreePozHashEntry(unsigned long entry)
 {if (entry>=PngWindowSize)
   {unsigned long indexValue=entry % PngWindowSize; //   -
    unsigned int oldHashValue=HashValue(hash_values[indexValue].index);
    if (countHashValue[oldHashValue]==1) //    
     {countHashValue[oldHashValue]=0;
      return;
     }
    countHashValue[oldHashValue]--;
   }
 }

 //   hash_table [hashvalue]  entry
void PngEncoder::MoveHashEntry (unsigned long entry, unsigned int hashvalue)
{ //   -
  unsigned long indexValue=entry % PngWindowSize; //   -
  hash_values[indexValue].next=hash_table[hashvalue];
  hash_values[indexValue].index=entry;
  hash_table[hashvalue]=&hash_values[indexValue];
  countHashValue[hashvalue]++;
}

//
//  Description:
//
//    This function finds the length encoding for
//    a length or distance table.
//
//  Parameters:
//    function: The function to process the code
//    lengths: lengths [n] = the length of the huffman code for n.
//    count: The number of coe lengths
//
void PngEncoder::FindLengthCodes (LENGTHFUNCTION function,
                                  UBYTE1 lengths [],
                                  unsigned int count)
{
  for (unsigned int ii = 0, jj ; ii < count ; ii = jj)
  {
    if (lengths [ii] != 0)
    { // 
      if (ii>0)
       if (lengths [ii] == lengths [ii-1])
        {for (jj = ii+1;
              lengths [jj] == lengths [ii-1] &&
              jj < count && jj-ii<6;
              ++ jj);
         if (jj-ii>2)
          {(this->*function) (ii, 16, 2, jj-ii-3);
           continue;
          }
        }
      // 
      (this->*function) (ii, lengths [ii], 0, 0) ;
      jj = ii + 1 ;
    }
    else
    {
      // Attempt to compact runs of zero length codes.
      // First find the number of consecutive zeros.
      for (jj = ii + 1;
           lengths [jj] == lengths [jj-1]
           && jj < count ;
           ++ jj)
      {
      }

      // We need at least 3 consecutive zeros to compact them.
      switch (jj - ii)
      {
      case 1:
        (this->*function) (ii, lengths [ii], 0, 0) ;
        break ;
      case 2:
        (this->*function) (ii, lengths [ii], 0, 0) ;
        (this->*function) (ii + 1, lengths [ii], 0, 0) ;
        break ;
      default:
        {
          // We have at least three zeros.

          int kk = jj - ii ;
          if (kk > 138)
          {
            kk = 138 ;
            jj = ii + kk ;
          }
          if (kk > 10)
          {
            (this->*function) (ii, 18, 7, kk - 11) ;
          }
          else
          {
            (this->*function) (ii, 17, 3, kk - 3) ;
          }
        }
        break ;
      }
    }
  }
  return ;
}


//
//  Description:
//
//    This function is passed as a parameter to FindLengthCodes.
//    It is use to find the frequency for each code.
//
//  Parameters:
//    code:  The code generated by FindLengthCodes
//
void PngEncoder::GatherLengthCounts (unsigned int,
                                     unsigned int code,
                                     unsigned int,
                                     unsigned int)
{
  length_length_table->IncrementFrequency (code) ;
  return ;
}

//
//  Description:
//
//    This function is passed as a parameter to FindLengthCodes.
//    It is use to encode and output the code to the output stream.
//
//  Parameters:
//    code:  The code generated by FindLengthCodes
//    extra:  The number of extra bits of data for the code
//    value:  The extra value
//
void PngEncoder::OutputLengthCounts (unsigned int,   // Index - Not used here
                                     unsigned int code,
                                     unsigned int extra,
                                     unsigned int value)
{
  UBYTE2 huffmancode ;
  UBYTE1 huffmansize ;
  length_length_table->Encode (code, huffmancode, huffmansize) ;

  OutputDataBits (huffmancode, huffmansize) ;
  if (extra != 0)
    OutputDataBits (value, extra) ;

  return ;
}

//
//  Description:
//
//    This function is called to start a new chunk.
//
//  Parameters:
//    type:  The PNG Chunk type (e.g. "IHDR", "IDAT")
//
void PngEncoder::StartChunk (const char type [5])
{
  chunk_type = ChunkType (type) ;
  bit_position = 0 ;
  byte_position = 0 ;
  output_buffer [0] = 0 ;
  return ;
}

//
//  Description:
//
//    This function outputs the current chunk.
//
void PngEncoder::WriteCurrentChunk ()
{
  // Calculate the CRC value for the chunk type and data
  UBYTE4 crc = 0xFFFFFFFFL ;
  crc = CRC32 (crc, &chunk_type, sizeof (chunk_type)) ;
  crc = CRC32 (crc, output_buffer, byte_position) ;
  crc ^= 0xFFFFFFFFL ;
  crc = SystemToBigEndian (crc) ;
  // Output the chunk. Length, type, data, CRC
  UBYTE4 length = SystemToBigEndian (byte_position) ;
  output_stream->write ((char *) &length, sizeof (length)) ;
  output_stream->write ((char *) &chunk_type, sizeof (chunk_type)) ;
  output_stream->write ((char *) output_buffer, byte_position) ;
  output_stream->write ((char *) &crc, sizeof (crc)) ;

  byte_position = 0 ;
  bit_position = 0 ;
  return ;
}

//
//  Description:
//
//    This function writes a block of data to the current chunk.
//
//  Parameters:
//    data: Pointer to the data to write
//    length: Number of bytes to write
//
void PngEncoder::OutputBlock (const void *data, unsigned int length)
{
  if (byte_position + length > OutputBufferSize)
    throw EPngError ("Output Buffer Overrun") ;

  memcpy (&output_buffer [byte_position], data, length) ;
  byte_position += length ;
  return;
}

//
//  Description:
//
//    This function writes the Adler value calculated for
//    the input data to the current chunk.
//    
void PngEncoder::OutputAdler ()
{
  FlushBitBuffer () ;
  adler_value = SystemToBigEndian (adler_value) ;
  UBYTE1 *buffer = (UBYTE1*) &adler_value ;
  for (unsigned int ii = 0 ; ii < sizeof (adler_value) ; ++ ii)
  {
    if (byte_position == OutputBufferSize)
      WriteCurrentChunk () ;

    output_buffer [byte_position] = buffer [ii] ;
    ++ byte_position ;
  }
  return ;
}

//
//  Description:
//
//    This function writes a number of data bits within an IDAT block.
//
//  Parameters:
//    data: The bits to output
//    length:  The number of bits to output
//
void PngEncoder::OutputDataBits (unsigned int data, unsigned int length)
{
  for (unsigned int ii = 0 ; ii < length ; ++ ii)
  {
    if (bit_position >= CHAR_BIT)
      FlushBitBuffer () ;

    if ((data & (1 << ii)) != 0)
    {
      output_buffer [byte_position] |= (1 << bit_position) ;
    }
    ++ bit_position ;
  }
  return ;
}

//
//  Description:
//
//    This function flushes any partial bytes that have been
//    written. This allows us to start writing byte data or 
//    start the next data byte.
//
void PngEncoder::FlushBitBuffer ()
{
  if (bit_position == 0)
    return ;

  ++ byte_position ;
  if (byte_position >= OutputBufferSize)
  {
    WriteCurrentChunk () ;
    StartChunk ("IDAT") ;
  }
  output_buffer [byte_position] = 0 ;
  bit_position = 0 ;
  return ;
}

//
//  Description:
//
//    This function encodes the image data. Depending upon
//    the function arguments this function either gathers
//    Huffman usage statistics or Huffman-encodes the data.
//
//  Return Value:
//    true => All data in the image has been processed
//    false => More data remains
//
bool PngEncoder::ProcessImageData ()
{
  // Here we loop until we either fill the block_buffer or
  // reach the end of the image.
  for (block_buffer_count = 0 ;
       block_buffer_count+countAnalizZamina < block_buffer_size - 2 //            
       && currentPozImage < nextPozStartBlockFilter[indexBlockFilter];
       ++ block_buffer_count)
  {
    unsigned int length ;
    unsigned int offset;
    unsigned int code, codeD;
    unsigned int extra;
    unsigned int value;
    LongestMatch (length, offset) ;
    if (length == 0)
    { unsigned int hashvalue = HashValue (currentPozImage) ;
      FreePozHashEntry(currentPozImage);
      MoveHashEntry (currentPozImage, hashvalue) ;

      UBYTE1 literal = imageDataPredict[currentPozImage];
      block_buffer [block_buffer_count] = literal ;
      length_table->IncrementFrequency (literal) ;
      adler_value = Adler (adler_value,imageData[currentPozImage]);
      currentPozImage++;
    }
    else
    { // Gather Huffman Statistics
      LengthToCode (length, code, extra, value) ;
	   length_table->IncrementFrequency (code) ;
      DistanceToCode (offset, codeD, extra, value) ;
	   distance_table->IncrementFrequency (codeD) ;

      if (trebaOptimizeLZBlock)
       {trebaZaminaBuffer[block_buffer_count]=1;  //     
        pozImageBuffer[block_buffer_count]=currentPozImage;
        pozAnalizZamina[countAnalizZamina]=block_buffer_count;    //     
        countAnalizZamina++;
       }
      // We have found a match.
      countZamina=countZamina+1;
      block_buffer [block_buffer_count] = 256 + length ;
      ++ block_buffer_count ;
      block_buffer [block_buffer_count] = offset ;
      // First update the hash table and
      // copy the data in the LZ window.
      for (unsigned int ii = 0 ; ii < length ; ii++)
      { FreePozHashEntry(currentPozImage);
        unsigned int hashvalue = HashValue (currentPozImage) ;
        MoveHashEntry (currentPozImage, hashvalue) ;
        adler_value = Adler (adler_value,imageData[currentPozImage]);
        currentPozImage++;
      }
    }
  }
  length_table->IncrementFrequency (PngEndCode) ;
  if (currentPozImage == countByteImage)
    return true ;
  else
   {while (currentPozImage >= nextPozStartBlockFilter[indexBlockFilter])
     indexBlockFilter++;
    return false ;
   }
}

//
//  Description:
//
//    This function outputs a single data byte.
//
//  Parameters:
//    value: The value to output
//
void PngEncoder::OutputByte (UBYTE1 value)
{
  output_buffer [byte_position] = value ;
  ++ byte_position ;
  return ;
}

//
//  Description:
//
//    This function calls the progress function if it has been
//    defined.
//
//  Parameters:
//    percent:  The completion percentage (0..100)
//
void PngEncoder::CallProgressFunction (unsigned int percent)
{
  if (progress_function == NULL)
    return ;

  if (percent > 100)
    percent = 100 ;

  bool cancel = false ;
  progress_function (*this, progress_data, 1, 1,
                     percent, cancel) ;
  return ;
}

//
//  Description:
// 
//    This function creates a tEXt chunk.
//
//  Parameters:
//    keyword:  The chunk keyword
//    value:  The keyword value
//
void PngEncoder::WriteText (const std::string &keyword,
                            const std::string &value)
{
  if (keyword.length () > 79)
    throw EPngError ("tEXt Keyword Too Long") ;
  if (value.length () + keyword.length () + 1 > OutputBufferSize)
    throw EPngError ("tEXt Value Too Long") ;

  StartChunk ("tEXt") ;
  OutputBlock (keyword.c_str (), keyword.length ()) ;
  OutputByte (0) ;
  OutputBlock (value.c_str (), value.length ()) ;
  WriteCurrentChunk () ;
  return ;
}

//
//  Description:
//
//    This function outputs tEXt blocks for any keywords
//    that have values assigned to them.
//
void PngEncoder::WriteTextBlocks ()
{
  if (title_string != "")
    WriteText ("Title", title_string) ;
  if (author_string != "")
    WriteText ("Author", author_string) ;
  if (description_string != "")
    WriteText ("Description", description_string) ;
  if (copyright_string != "")
    WriteText ("Copyright", copyright_string) ;
  if (software_string != "")
    WriteText ("Software", software_string) ;
  if (disclaimer_string != "")
    WriteText ("Disclaimer", disclaimer_string) ;
  if (warning_string != "")
    WriteText ("Warning", warning_string) ;
  if (source_string != "")
    WriteText ("Source", source_string) ;
  if (comment_string != "")
    WriteText ("Comment", comment_string) ;
  return ;
}

//
//  Description:
//
//    This function sets the compression level used.  The compression level
//    determines the depth to which hash chains are searched.
//
//  Parameters:
//    value:  The new compression level
//
void PngEncoder::SetCompressionLevel (PngEncoder::CompressionLevel value)
{
  if (value < 0 || value > 3)
    throw EPngError ("Invalid Compression Level") ;

  compression_level = value ;
  return ;
}

//
//  Description:
//
//    This function performs the actual output of the image.
//

double log2(double x)
 {return log(x)*1.4426950409;

 }

double entropiCode(long * freq, short countAllFreq=256)
 {// freq -   
  // countAllFreq -     
  double len=0, entropi; long n=0;
  for (long i=0; i<countAllFreq; i++)
   {n+=freq[i]; //  
    if (freq[i]>1) //   log
     len+=freq[i]*log(freq[i]); }
  if (n) {entropi=(n*log(n)-len)/(n*log(2)); }
  else {entropi=0; }
  return entropi;
 }

double lenEntropiCode(long * freq, short countAllFreq=256)
 {// freq -   
  // countAllFreq -     
  double len=0, lenEntropi; long n=0;
  for (long i=0; i<countAllFreq; i++)
   {n+=freq[i]; //  
    if (freq[i]>1) //   log
     len+=freq[i]*log(freq[i]); }
  if (n) {lenEntropi=(n*log(n)-len)/log(2); }
  else {lenEntropi=0; }
  return lenEntropi;
 }

UBYTE1 ProgresPredict1(UBYTE1 a, UBYTE1 b,
                       UBYTE1 c, UBYTE1 d)
 {UBYTE1 pa, pb;
  if (a>=c)
   pa=a-c;
  else
   pa=c-a;
  if (b>=d)
   pb=b-d;
  else
   pb=d-b;
  if (pa<=pb)
   return (a+c)/2;
  else
   return (b+d)/2;
 }

UBYTE1 ProgresPredict2(UBYTE1 a, UBYTE1 b, UBYTE1 c, UBYTE1 d)
 {UBYTE1 absac, absbd, maxac, minac, maxbd, minbd, prognozn;
  if (a<=c)
   {absac=c-a;
    maxac=c;
    minac=a;
   }
  else
   {absac=a-c;
    maxac=a;
    minac=c;
   }
  if (b<=d)
   {absbd=d-b;
    maxbd=d;
    minbd=b;
   }
  else
   {absbd=b-d;
    maxbd=b;
    minbd=d;
   }
  if (absac<=absbd)
   {if (minbd>=minac && minbd<=maxac)
     return minbd;
    if (maxbd>=minac && maxbd<=maxac)
     return maxbd;
    return minac;
   }
  else
   {if (minac>=minbd && minac<=maxbd)
     return minac;
    if (maxac>=minbd && maxac<=maxbd)
     return maxac;
    return minbd;
   }
 }

UBYTE1 ProgresPredict4(UBYTE1 a, UBYTE1 b, UBYTE1 c, UBYTE1 d, UBYTE1 ad, UBYTE1 ab)
 {UBYTE1 absac, absbd, maxac, minac, maxbd, minbd, prognozn, s, mins;
  if (a<=c)
   {absac=c-a;
    maxac=c;
    minac=a;
   }
  else
   {absac=a-c;
    maxac=a;
    minac=c;
   }
  if (b<=d)
   {absbd=d-b;
    maxbd=d;
    minbd=b;
   }
  else
   {absbd=b-d;
    maxbd=b;
    minbd=d;
   }
  if (absac<=absbd)
   {prognozn=ad;
    mins=abs(maxac+minac-2*ad);
    s=abs(maxac+minac-2*ab);
    if (s<mins)
     {prognozn=ab;
      mins=s;
     }
    s=abs(maxac+minac-2*minbd);
    if (s<mins)
     {prognozn=minbd;
      mins=s;
     }
    s=abs(maxac+minac-2*maxbd);
    if (s<mins)
     {prognozn=maxbd;
      mins=s;
     }
    if (minac>=prognozn || prognozn>maxac)
     return minac;
    return prognozn;
   }
  else
   {prognozn=ad;
    mins=abs(maxbd+minbd-2*ad);
    s=abs(maxbd+minbd-2*ab);
    if (s<mins)
     {prognozn=ab;
      mins=s;
     }
    s=abs(maxbd+minbd-2*minac);
    if (s<mins)
     {prognozn=minac;
      mins=s;
     }
    s=abs(maxbd+minbd-2*maxac);
    if (s<mins)
     {prognozn=maxac;
      mins=s;
     }
    if (minbd>=prognozn || prognozn>maxbd)
     return minbd;
    return prognozn;
   }
 }

UBYTE1 ProgresPredict5(UBYTE1 a, UBYTE1 b, UBYTE1 c, UBYTE1 d, UBYTE1 ad, UBYTE1 ab)
 {UBYTE1 absac, absbd, maxac, minac, maxbd, minbd, prognozn, minprD, minprGV, prD, prGV;
  if (a<=c)
   {absac=c-a;
    maxac=c;
    minac=a;
   }
  else
   {absac=a-c;
    maxac=a;
    minac=c;
   }
  if (b<=d)
   {absbd=d-b;
    maxbd=d;
    minbd=b;
   }
  else
   {absbd=b-d;
    maxbd=b;
    minbd=d;
   }
  minprD=abs(ab-a);
  minprGV=abs(a-d);
  prognozn=d;
  prD=abs(ab-b);
  if (prD<=minprD)
   {prGV=abs(b-c);
    if (prD<minprD || prGV<minprGV)
     {minprD=prD;
      minprGV=prGV;
      prognozn=c;
     }
   }
  prD=abs(ad-a);
  if (prD<=minprD)
   {prGV=abs(a-b);
    if (prD<minprD || prGV<minprGV)
     {minprD=prD;
      minprGV=prGV;
      prognozn=b;
     }
   }
  prD=abs(ad-d);
  if (prD<=minprD)
   {prGV=abs(d-c);
    if (prD<minprD || prGV<minprGV)
     {minprD=prD;
      minprGV=prGV;
      prognozn=c;
     }
   }
  if (2*minprD<min(absac,absbd))
   return prognozn;
  if (absac<=absbd)
   {if (minbd>=minac && minbd<=maxac)
     return minbd;
    if (maxbd>=minac && maxbd<=maxac)
     return maxbd;
    return minac;
   }
  else
   {if (minac>=minbd && minac<=maxbd)
     return minac;
    if (maxac>=minbd && maxac<=maxbd)
     return maxac;
    return minbd;
   }
 }

double sizeEntropiCode(UBYTE4 *masFreq, unsigned int countAllFreq=256)
 {UBYTE4 count=0, i;

  double size=0;

  for (i=0; i<countAllFreq; i++)

   {count+=masFreq[i];

    if (masFreq[i]>1)

     size+=masFreq[i]*log(masFreq[i]);

   }

  if (count)

   size=(count*log(count)-size)/log(2);

  return size;

 }

void PngEncoder::initCM() //      
 {//   
  UBYTE2 m, n, r;
  int i, j, k, l;
  unsigned int step;
  UBYTE4 freq[256];
  UBYTE4 prognozLen[3][3], minusLen;
  UBYTE1 Left, Above, Right, Down, LeftAbove, LeftDown,RightAbove, RightDown;
  k=0;
  if (min(image_col,image_row)>15)
   k=log2((min(image_col,image_row)-1)/15);
  for (m=0; m<=2; m++) //  
   for (n=0; n<=2; n++) //   
    if (m==n) //     
     {memset(freq, 0, sizeof(freq));
      step=1;
      for (l=k+1; l>=2; l--) //  
       {step*=2;
        //        
        for (j=0; j<image_row; j+=step)
         for (i=step/2; i<image_col; i+=step)
          {if (j>=step/2)
            Above=(*current_image)[j-step/2][3*i+m];
           else
            Above=0;
           if (j+step/2<image_row)
            Down=(*current_image)[j+step/2][3*i+m];
           else
            Down=0;
           if (i>=step/2)
            Left=(*current_image)[j][3*(i-step/2)+m];
           else
            Left=0;
           if (i+step/2<image_col)
            Right=(*current_image)[j][3*(i+step/2)+m];
           else
            Right=0;
           freq[(byte)((*current_image)[j][3*i+m]-ProgresPredict1(Left, Above, Right, Down))]++;
          }
        //        
        for (j=step/2; j<image_row; j+=step)
         for (i=0; i<image_col; i+=step)
          {if (j>=step/2)
            Above=(*current_image)[j-step/2][3*i+m];
           else
            Above=0;
           if (j+step/2<image_row)
            Down=(*current_image)[j+step/2][3*i+m];
           else
            Down=0;
           if (i>=step/2)
            Left=(*current_image)[j][3*(i-step/2)+m];
           else
            Left=0;
           if (i+step/2<image_col)
            Right=(*current_image)[j][3*(i+step/2)+m];
           else
            Right=0;
           freq[(byte)((*current_image)[j][3*i+m]-ProgresPredict1(Left, Above, Right, Down))]++;
          }
        //      
        for (j=step/2; j<image_row; j+=step)
         for (i=step/2; i<image_col; i+=step)
          {LeftAbove=(*current_image)[j-step/2][3*(i-step/2)+m];
           if (j+step/2<image_row)
            LeftDown=(*current_image)[j+step/2][3*(i-step/2)+m];
           else
            LeftDown=0;
           if (i+step/2<image_col)
            RightAbove=(*current_image)[j-step/2][3*(i+step/2)+m];
           else
            RightAbove=0;
           if (j+step/2<image_row && i+step/2<image_col)
            RightDown=(*current_image)[j+step/2][3*(i+step/2)+m];
           else
            RightDown=0;
           freq[(byte)((*current_image)[j][3*i+m]-ProgresPredict1(LeftAbove, RightAbove, RightDown, LeftDown))]++;
          }
       }
      prognozLen[m][m]=sizeEntropiCode(freq);
     }
    else
     if (n>m)
      {memset(freq, 0, sizeof(freq));
       step=1;
       for (l=k+1; l>=2; l--) //  
        {step*=2;
         //        
         for (j=0; j<image_row; j+=step)
          for (i=step/2; i<image_col; i+=step)
           {if (j>=step/2)
             Above=(*current_image)[j-step/2][3*i+m]-(*current_image)[j-step/2][3*i+n];
            else
             Above=0;
            if (j+step/2<image_row)
             Down=(*current_image)[j+step/2][3*i+m]-(*current_image)[j+step/2][3*i+n];
            else
             Down=0;
            if (i>=step/2)
             Left=(*current_image)[j][3*(i-step/2)+m]-(*current_image)[j][3*(i-step/2)+n];
            else
             Left=0;
            if (i+step/2<image_col)
             Right=(*current_image)[j][3*(i+step/2)+m]-(*current_image)[j][3*(i+step/2)+n];
            else
             Right=0;
            freq[(byte)((byte)((*current_image)[j][3*i+m]-(*current_image)[j][3*i+n])-
                 ProgresPredict1(Left, Above, Right, Down))]++;
           }
         //        
         for (j=step/2; j<image_row; j+=step)
          for (i=0; i<image_col; i+=step)
           {if (j>=step/2)
             Above=(*current_image)[j-step/2][3*i+m]-(*current_image)[j-step/2][3*i+n];
            else
             Above=0;
            if (j+step/2<image_row)
             Down=(*current_image)[j+step/2][3*i+m]-(*current_image)[j+step/2][3*i+n];
            else
             Down=0;
            if (i>=step/2)
             Left=(*current_image)[j][3*(i-step/2)+m]-(*current_image)[j][3*(i-step/2)+n];
            else
             Left=0;
            if (i+step/2<image_col)
             Right=(*current_image)[j][3*(i+step/2)+m]-(*current_image)[j][3*(i+step/2)+n];
            else
             Right=0;
            freq[(byte)((byte)((*current_image)[j][3*i+m]-(*current_image)[j][3*i+n])-
                 ProgresPredict1(Left, Above, Right, Down))]++;
           }
         //      
         for (j=step/2; j<image_row; j+=step)
          for (i=step/2; i<image_col; i+=step)
           {LeftAbove=(*current_image)[j-step/2][3*(i-step/2)+m]-(*current_image)[j-step/2][3*(i-step/2)+n];
            if (j+step/2<image_row)
             LeftDown=(*current_image)[j+step/2][3*(i-step/2)+m]-(*current_image)[j+step/2][3*(i-step/2)+n];
            else
             LeftDown=0;
            if (i+step/2<image_col)
             RightAbove=(*current_image)[j-step/2][3*(i+step/2)+m]-(*current_image)[j-step/2][3*(i+step/2)+n];
            else
             RightAbove=0;
            if (j+step/2<image_row && i+step/2<image_col)
             RightDown=(*current_image)[j+step/2][3*(i+step/2)+m]-(*current_image)[j+step/2][3*(i+step/2)+n];
            else
             RightDown=0;
            freq[(byte)((byte)((*current_image)[j][3*i+m]-(*current_image)[j][3*i+n])-
                 ProgresPredict1(LeftAbove, RightAbove, RightDown, LeftDown))]++;
           }
        }
       prognozLen[m][n]=sizeEntropiCode(freq);
      }
     else
      prognozLen[m][n]=prognozLen[n][m];
  UBYTE1 c11, c12, c21, c22;
  cm11=0, cm12=0, cm21=0, cm22=0;
  minusLen=0;
  for (c11=0; c11<=2; c11++)
   for (c12=0; c12<=2; c12++)
    if (c11!=c12 && prognozLen[c11][c12]<prognozLen[c11][c11]) //     
     {if (prognozLen[c11][c11]-prognozLen[c11][c12]>minusLen)
       {minusLen=prognozLen[c11][c11]-prognozLen[c11][c12];
        cm11=c11;
        cm12=c12;
        cm21=0;
        cm22=0;
       }
      for (c21=c11+1; c21<=2; c21++)
       for (c22=0; c22<=2; c22++)
        if (c21!=c22 && (c21!=c12 || c22!=c11) && prognozLen[c21][c22]<prognozLen[c21][c21]) //     
         if (prognozLen[c11][c11]-prognozLen[c11][c12]+
             prognozLen[c21][c21]-prognozLen[c21][c22]>minusLen)
          {minusLen=prognozLen[c11][c11]-prognozLen[c11][c12]+
                    prognozLen[c21][c21]-prognozLen[c21][c22];
           cm11=c11;
           cm12=c12;
           cm21=c21;
           cm22=c22;
          }
     }
  if (minusLen<(prognozLen[0][0]+prognozLen[1][1]+prognozLen[2][2])/100)
   {cm11=0;
    cm12=0;
    cm21=0;
    cm22=0;
    return;
   }
  if (cm11==cm22) //   
   {i=cm11;
    cm11=cm21;
    cm21=i;
    i=cm12;
    cm12=cm22;
    cm22=i;
   }
  if (cm11!=cm12) //  
   for (i = 0 ; i < image_row; i++)
    for (r = 0; r<row_width; r+=3)
     (*current_image) [i][r+cm11]-=(*current_image)[i][r+cm12];
  if (cm21!=cm22) //  
   for (i = 0 ; i < image_row; i++)
    for (r = 0; r<row_width; r+=3)
     (*current_image) [i][r+cm21]-=(*current_image)[i][r+cm22];
 }


void PngEncoder::doneCM()
 {UBYTE2 i, r;
  if (cm21!=cm22) //  
   {for (i = 0; i < image_row; i++)
     for (r = 0; r<row_width; r+=3)
      (*current_image)[i][r+cm21]+=(*current_image)[i][r+cm22];
    cm21=cm22=0;
   }
  if (cm11!=cm12) //  
   {for (i = 0; i < image_row; i++)
     for (r = 0; r<row_width; r+=3)
      (*current_image)[i][r+cm11]+=(*current_image)[i][r+cm12];
    cm11=cm12=0;
   }
 }

void PngEncoder::writeCM()
 {if (cm11!=cm12)
   {StartChunk ("cmOd");
    UBYTE1 cm=cm11<<6 | cm12<<4 | cm21<<2 |cm22;
    OutputBlock (&cm, 1) ;
    WriteCurrentChunk () ;
   }
 }



void PngEncoder::DoWrite ()
{
  initCodesDistance();
  // Needed because MSVC++ does not follow standard scoping rules
  // in for statements.
  unsigned int ii ;

  if (current_image->BitCount () == 24)
  {
    row_width = 3 * current_image->Width () ;
    countComponent = 3 ;
  }
  else
  {
    row_width = (current_image->BitCount () * current_image->Width ()
                 + CHAR_BIT - 1) / CHAR_BIT ;
    countComponent = 1 ;
  }

  int i, j, k, l;
  for (j = 0 ; j < FilterBufferCount ; ++ j)
   if (((1 << j) & filter_mask) != 0)
    filter_buffers [j] = new UBYTE1 [row_width] ;

  image_row=current_image->Height();
  image_col=current_image->Width();
  imageData=new UBYTE1[image_row*(row_width+1)+500];
  imageDataPredict=new UBYTE1[image_row*(row_width+1)];
  PngWindowSize= (1<<15) ; //*3; // -  PNG-
  PngWindowSize-=PngWindowSize%3;

  hash_values = new HashEntry [PngWindowSize] ;
  hash_table = new PHashEntry[HashTableSize] ;
  countHashValue = new unsigned long[HashTableSize];
  output_buffer = new UBYTE1 [OutputBufferSize] ;

  switch (compression_level)
   {case FastestCompression:
     search_limit = 1 ;
     break;
    case FastCompression:
     search_limit = 64 ;
     break ;
    case DefaultCompression:
     search_limit = 128 ;
     break ;
    case MaximumCompression:
     search_limit = UINT_MAX ;
     break ;
    default:
     throw EPngError ("Invalid Compression Level") ;
   }

  SYSTEMTIME startTimeAnaliz, endTimeAnaliz;
  GetLocalTime(&startTimeAnaliz);
  cm11=0, cm12=0, cm21=0, cm22=0;
  if (trebaPerexidCM)
   {initCM();
    GetLocalTime(&endTimeAnaliz);
    long int intervalAnaliz= (endTimeAnaliz.wHour-startTimeAnaliz.wHour)*60*60*1000 +
                       (endTimeAnaliz.wMinute-startTimeAnaliz.wMinute)*60*1000 +
                       (endTimeAnaliz.wSecond-startTimeAnaliz.wSecond)*1000 +
                       (endTimeAnaliz.wMilliseconds-startTimeAnaliz.wMilliseconds);
    cout << endl << "perexid CM " << intervalAnaliz << " milliseconds " ;
    cout << endl << "parametr CM (model BGR) " << (int)cm11 << " " << (int)cm12 << " " << (int)cm21 << " " << (int)cm22;
   }
  memset(imageData,0,image_row*(row_width+1)+500);
  for (j=0; j<image_row;j++)
   {FilterRow(j);
    imageData[j*(row_width+1)]=current_filter;
    for (i=0; i<image_col;i++)
     {imageData[j*(row_width+1)+1+3*i]=filter_buffers [current_filter][3*i+2];
      imageData[j*(row_width+1)+1+3*i+1]=filter_buffers [current_filter][3*i+1];
      imageData[j*(row_width+1)+1+3*i+2]=filter_buffers [current_filter][3*i];
     }
   }
  memcpy(imageDataPredict, imageData, image_row*(row_width+1));
  unsigned int step=1, stepShar1, indexDelta, index;
  UBYTE1 * deltaPredict1, * deltaPredict2, * pprRow, * deltaPredict4, * deltaPredict5;
  deltaPredict1=new UBYTE1[row_width];
  deltaPredict2=new UBYTE1[row_width];
  deltaPredict4=new UBYTE1[row_width];
  deltaPredict5=new UBYTE1[row_width];
  pprRow=new UBYTE1[row_width];
  UBYTE1 Above, Left, Right, Down, LeftAbove, RightAbove, LeftDown, RightDown, LeftLeft, AboveAbove, predict;
  UBYTE1 component, pprNomerPredict, nomerPredict;
  long freq[256], freq1[256], freq2[256], freq3[256], freq4[256], freq5[256];
  double entropi, entropi1, entropi2, entropi3, entropi4, entropi5, lenEntropi=0, lenEntropi1, lenEntropi2, lenEntropiN;
  bool trebaPredictP1, trebaPredictP21, trebaPredictP22;
  k=0;
  if (min(image_col,image_row)>15)
   k=log2((min(image_col,image_row)-1)/15);
  for (l=k+1; l>=2; l--) //  
   {step*=2;
    //        
    memset(freq,0,256*sizeof(long));
    memset(freq1,0,256*sizeof(long));
    memset(freq2,0,256*sizeof(long));
    for (j=0; j<image_row; j+=step)
     for (i=step/2; i<image_col; i+=step)
      for (component=0; component<=2; component++)
       {if (j>=step/2)
         Above=imageData[(j-step/2)*(row_width+1)+1+3*i+component];
        else
         Above=0;
        if (j+step/2<image_row)
         Down=imageData[(j+step/2)*(row_width+1)+1+3*i+component];
        else
         Down=0;
        if (i>=step/2)
         Left=imageData[j*(row_width+1)+1+3*(i-step/2)+component];
        else
         Left=0;
        if (i+step/2<image_col)
         Right=imageData[j*(row_width+1)+1+3*(i+step/2)+component];
        else
         Right=0;
        freq[imageData[j*(row_width+1)+1+3*i+component]]++; //  
        predict=ProgresPredict1(Left, Above, Right, Down);
        freq1[(UBYTE1)(imageData[j*(row_width+1)+1+3*i+component]-predict)]++; //   
        predict=ProgresPredict2(Left, Above, Right, Down);
        freq2[(UBYTE1)(imageData[j*(row_width+1)+1+3*i+component]-predict)]++;
       }
    lenEntropiN=lenEntropiCode(freq, 256);
    lenEntropi1=lenEntropiCode(freq1, 256);
    lenEntropi2=lenEntropiCode(freq2, 256);
    if (lenEntropi1<lenEntropiN || lenEntropi2<lenEntropiN) //   
     trebaPredictP22=true;
    else
     trebaPredictP22=false;
    //        
    memset(freq,0,256*sizeof(long));
    memset(freq1,0,256*sizeof(long));
    memset(freq2,0,256*sizeof(long));
    for (j=step/2; j<image_row; j+=step)
     for (i=0; i<image_col; i+=step)
      for (component=0; component<=2; component++)
       {if (j>=step/2)
         Above=imageData[(j-step/2)*(row_width+1)+1+3*i+component];
        else
         Above=0;
        if (j+step/2<image_row)
         Down=imageData[(j+step/2)*(row_width+1)+1+3*i+component];
        else
         Down=0;
        if (i>=step/2)
         Left=imageData[j*(row_width+1)+1+3*(i-step/2)+component];
        else
         Left=0;
        if (i+step/2<image_col)
         Right=imageData[j*(row_width+1)+1+3*(i+step/2)+component];
        else
         Right=0;
        freq[imageData[j*(row_width+1)+1+3*i+component]]++; //  
        predict=ProgresPredict1(Left, Above, Right, Down);
        freq1[(UBYTE1)(imageData[j*(row_width+1)+1+3*i+component]-predict)]++; //   
        predict=ProgresPredict2(Left, Above, Right, Down);
        freq2[(UBYTE1)(imageData[j*(row_width+1)+1+3*i+component]-predict)]++;
       }
    lenEntropiN=lenEntropiCode(freq, 256);
    lenEntropi1=lenEntropiCode(freq1, 256);
    lenEntropi2=lenEntropiCode(freq2, 256);
    if (lenEntropi1<lenEntropiN || lenEntropi2<lenEntropiN ) //   
     trebaPredictP21=true;
    else
     trebaPredictP21=false;
    //      
    memset(freq,0,256*sizeof(long));
    memset(freq1,0,256*sizeof(long));
    memset(freq2,0,256*sizeof(long));
    for (j=step/2; j<image_row; j+=step)
     for (i=step/2; i<image_col; i+=step)
      for (component=0; component<=2; component++)
       {LeftAbove=imageData[(j-step/2)*(row_width+1)+1+3*(i-step/2)+component];
        if (j+step/2<image_row)
         LeftDown=imageData[(j+step/2)*(row_width+1)+1+3*(i-step/2)+component];
        else
         LeftDown=0;
        if (i+step/2<image_col)
         RightAbove=imageData[(j-step/2)*(row_width+1)+1+3*(i+step/2)+component];
        else
         RightAbove=0;
        if (j+step/2<image_row && i+step/2<image_col)
         RightDown=imageData[(j+step/2)*(row_width+1)+1+3*(i+step/2)+component];
        else
         RightDown=0;
        freq[imageData[j*(row_width+1)+1+3*i+component]]++; //  
        predict=ProgresPredict1(LeftAbove, RightAbove, RightDown, LeftDown);
        freq1[(UBYTE1)(imageData[j*(row_width+1)+1+3*i+component]-predict)]++; //   
        predict=ProgresPredict2(LeftAbove, RightAbove, RightDown, LeftDown);
        freq2[(UBYTE1)(imageData[j*(row_width+1)+1+3*i+component]-predict)]++;
       }
    lenEntropiN=lenEntropiCode(freq, 256);
    lenEntropi1=lenEntropiCode(freq1, 256);
    lenEntropi2=lenEntropiCode(freq2, 256);
    if (lenEntropi1<lenEntropiN || lenEntropi2<lenEntropiN) //   
     trebaPredictP1=true;
    else
     trebaPredictP1=false;

    memset(freq,0,256*sizeof(long)); //      
    if (trebaPredictP22)
     {pprNomerPredict=3; //        
      //     
      for (j=0; j<image_row; j+=step)
       {memset(freq1,0,256*sizeof(long));
        memset(freq2,0,256*sizeof(long));
        indexDelta=0;
        for (i=step/2; i<image_col; i+=step)
         for (component=0; component<=2; component++)
          {if (j>=step/2)
            Above=imageData[(j-step/2)*(row_width+1)+1+3*i+component];
           else
            Above=0;
           if (j+step/2<image_row)
            Down=imageData[(j+step/2)*(row_width+1)+1+3*i+component];
           else
            Down=0;
           if (i>=step/2)
            Left=imageData[j*(row_width+1)+1+3*(i-step/2)+component];
           else
            Left=0;
           if (i+step/2<image_col)
            Right=imageData[j*(row_width+1)+1+3*(i+step/2)+component];
           else
            Right=0;
           predict=ProgresPredict1(Left, Above, Right, Down);
           deltaPredict1[indexDelta]=imageData[j*(row_width+1)+1+3*i+component]-predict;
           freq1[deltaPredict1[indexDelta]]++; //  -uj 
           predict=ProgresPredict2(Left, Above, Right, Down);
           deltaPredict2[indexDelta]=imageData[j*(row_width+1)+1+3*i+component]-predict;
           freq2[deltaPredict2[indexDelta]]++; //  - 
           indexDelta++;
          }
        if (j+step/2<image_row) //      
         {entropi1=entropiCode(freq1, 256);
          entropi2=entropiCode(freq2, 256);
          if (entropi2<entropi1)
           nomerPredict=2;
          else
           nomerPredict=1;
          if (nomerPredict!=pprNomerPredict)
           {imageData[(j+step/2)*(row_width+1)]=imageData[(j+step/2)*(row_width+1)] | (nomerPredict<<6);
            imageDataPredict[(j+step/2)*(row_width+1)]=imageData[(j+step/2)*(row_width+1)];
            pprNomerPredict=nomerPredict;
           }
         }
        else
         nomerPredict=pprNomerPredict;
        //  
        indexDelta=0;
        for (i=step/2; i<image_col; i+=step)
         for (component=0; component<=2; component++)
          {if (nomerPredict==1)
            imageDataPredict[j*(row_width+1)+1+3*i+component]=deltaPredict1[indexDelta];
           else
            imageDataPredict[j*(row_width+1)+1+3*i+component]=deltaPredict2[indexDelta];
           indexDelta++;
           freq[imageDataPredict[j*(row_width+1)+1+3*i+component]]++;
          }
       }
     }
    else
     //     
     for (j=0; j<image_row; j+=step)
      for (i=step/2; i<image_col; i+=step)
       for (component=0; component<=2; component++)
        freq[imageData[j*(row_width+1)+1+3*i+component]]++;
    entropi=entropiCode(freq, 256);
    lenEntropi+=lenEntropiCode(freq, 256);
    cout << endl << step <<" - entropi "<< entropi << " bpb";

    memset(freq,0,256*sizeof(long));
    if (trebaPredictP21)
     {pprNomerPredict=3; //        
      //     
      for (j=step/2; j<image_row; j+=step)
       {memset(freq1,0,256*sizeof(long));
        memset(freq2,0,256*sizeof(long));
        indexDelta=0;
        for (i=0; i<image_col; i+=step)
         for (component=0; component<=2; component++)
          {if (j>=step/2)
            Above=imageData[(j-step/2)*(row_width+1)+1+3*i+component];
           else
            Above=0;
           if (j+step/2<image_row)
            Down=imageData[(j+step/2)*(row_width+1)+1+3*i+component];
           else
            Down=0;
           if (i>=step/2)
            Left=imageData[j*(row_width+1)+1+3*(i-step/2)+component];
           else
            Left=0;
           if (i+step/2<image_col)
            Right=imageData[j*(row_width+1)+1+3*(i+step/2)+component];
           else
            Right=0;
           predict=ProgresPredict1(Left, Above, Right, Down);
           deltaPredict1[indexDelta]=imageData[j*(row_width+1)+1+3*i+component]-predict;
           freq1[deltaPredict1[indexDelta]]++;
           predict=ProgresPredict2(Left, Above, Right, Down);
           deltaPredict2[indexDelta]=imageData[j*(row_width+1)+1+3*i+component]-predict;
           freq2[deltaPredict2[indexDelta]]++;
           indexDelta++;
          }
        entropi1=entropiCode(freq1, 256);
        entropi2=entropiCode(freq2, 256);
        if (entropi2<entropi1)
         nomerPredict=2;
        else
         nomerPredict=1;
        if (nomerPredict!=pprNomerPredict)
         {imageData[j*(row_width+1)]=imageData[j*(row_width+1)] | (nomerPredict<<4);
          imageDataPredict[j*(row_width+1)]=imageData[j*(row_width+1)];
          pprNomerPredict=nomerPredict;
         }
        //  
        indexDelta=0;
        for (i=0; i<image_col; i+=step)
         for (component=0; component<=2; component++)
          {if (nomerPredict==1)
            imageDataPredict[j*(row_width+1)+1+3*i+component]=deltaPredict1[indexDelta];
           else
            imageDataPredict[j*(row_width+1)+1+3*i+component]=deltaPredict2[indexDelta];
           indexDelta++;
           freq[imageDataPredict[j*(row_width+1)+1+3*i+component]]++;
          }
       }
     }
    else
     //     
     for (j=step/2; j<image_row; j+=step)
      for (i=0; i<image_col; i+=step)
       for (component=0; component<=2; component++)
        freq[imageData[j*(row_width+1)+1+3*i+component]]++;
    entropi=entropiCode(freq, 256);
    lenEntropi+=lenEntropiCode(freq, 256);
    cout << endl << step <<" - entropi "<< entropi << " bpb";

    memset(freq,0,256*sizeof(long));
    if (trebaPredictP1)
     {pprNomerPredict=3; //        
      memset(pprRow,0,row_width); //    
      //    
      for (j=step/2; j<image_row; j+=step)
       {memset(freq1,0,256*sizeof(long));
        memset(freq2,0,256*sizeof(long));
        memset(freq4,0,256*sizeof(long));
        memset(freq5,0,256*sizeof(long));
        indexDelta=0;
        for (i=step/2; i<image_col; i+=step)
         for (component=0; component<=2; component++)
          {LeftAbove=imageData[(j-step/2)*(row_width+1)+1+3*(i-step/2)+component];
           if (j+step/2<image_row)
            LeftDown=imageData[(j+step/2)*(row_width+1)+1+3*(i-step/2)+component];
           else
            LeftDown=0;
           if (i+step/2<image_col)
            RightAbove=imageData[(j-step/2)*(row_width+1)+1+3*(i+step/2)+component];
           else
            RightAbove=0;
           if (j+step/2<image_row && i+step/2<image_col)
            RightDown=imageData[(j+step/2)*(row_width+1)+1+3*(i+step/2)+component];
           else
            RightDown=0;
           if (i>=step)
            LeftLeft=imageData[j*(row_width+1)+1+3*(i-step)+component];
           else
            LeftLeft=0;
           AboveAbove=pprRow[3*i+component];
           predict=ProgresPredict1(LeftAbove, RightAbove, RightDown, LeftDown);
           deltaPredict1[indexDelta]=imageData[j*(row_width+1)+1+3*i+component]-predict;
           freq1[deltaPredict1[indexDelta]]++;
           predict=ProgresPredict2(LeftAbove, RightAbove, RightDown, LeftDown);
           deltaPredict2[indexDelta]=imageData[j*(row_width+1)+1+3*i+component]-predict;
           freq2[deltaPredict2[indexDelta]]++;
           predict=ProgresPredict4(LeftAbove, RightAbove, RightDown, LeftDown, LeftLeft, AboveAbove);
           deltaPredict4[indexDelta]=imageData[j*(row_width+1)+1+3*i+component]-predict;
           freq4[deltaPredict4[indexDelta]]++;
           predict=ProgresPredict5(LeftAbove, RightAbove, RightDown, LeftDown, LeftLeft, AboveAbove);
           deltaPredict5[indexDelta]=imageData[j*(row_width+1)+1+3*i+component]-predict;
           freq5[deltaPredict5[indexDelta]]++;
           indexDelta++;
          }
        entropi1=entropiCode(freq1, 256);
        entropi2=entropiCode(freq2, 256);
        entropi4=entropiCode(freq4, 256);
        entropi5=entropiCode(freq5, 256);
        if (entropi1<=entropi2 && entropi1<=entropi4 && entropi1<=entropi5)
         nomerPredict=1;
        else
         if (entropi2<=entropi4 && entropi2<=entropi5)
          nomerPredict=2;
         else
          if (entropi4<=entropi5)
           nomerPredict=4;
          else
           nomerPredict=5;
        if (nomerPredict!=pprNomerPredict)
         {imageData[j*(row_width+1)]=imageData[j*(row_width+1)] | nomerPredict;
          imageDataPredict[j*(row_width+1)]=imageData[j*(row_width+1)];
          pprNomerPredict=nomerPredict;
         }
        //     
        memcpy(pprRow,imageData+j*(row_width+1)+1,row_width);
        //  
        indexDelta=0;
        for (i=step/2; i<image_col; i+=step)
         for (component=0; component<=2; component++)
          {switch (nomerPredict)
            {case 1:imageDataPredict[j*(row_width+1)+1+3*i+component]=deltaPredict1[indexDelta];
                    break;
             case 2:imageDataPredict[j*(row_width+1)+1+3*i+component]=deltaPredict2[indexDelta];
                    break;
             case 4:imageDataPredict[j*(row_width+1)+1+3*i+component]=deltaPredict4[indexDelta];
                    break;
             case 5:imageDataPredict[j*(row_width+1)+1+3*i+component]=deltaPredict5[indexDelta];
                    break;
            }
           indexDelta++;
           freq[imageDataPredict[j*(row_width+1)+1+3*i+component]]++;
          }
       }
     }
    else
     //     
     for (j=step/2; j<image_row; j+=step)
      for (i=step/2; i<image_col; i+=step)
       for (component=0; component<=2; component++)
        freq[imageData[j*(row_width+1)+1+3*i+component]]++;
    entropi=entropiCode(freq, 256);
    lenEntropi+=lenEntropiCode(freq, 256);
    cout << endl << step <<" - entropi "<< entropi << " bpb";
   }
  //    ϳ   
  memset(freq,0,256*sizeof(long));
  memset(freq1,0,256*sizeof(long));
  for (j=0; j<image_row; j+=step)
   for (i=0; i<image_col; i+=step)
    for (component=0; component<=2; component++)
     {if (j>=step)
       Above=imageData[(j-step)*(row_width+1)+1+3*i+component];
      else
       Above=0;
      if (j>=step && i>=step)
       LeftAbove=imageData[(j-step)*(row_width+1)+1+3*(i-step)+component];
      else
       LeftAbove=0;
      if (i>=step)
       Left=imageData[j*(row_width+1)+1+3*(i-step)+component];
      else
       Left=0;
      predict=PaethPredictor(Left, Above, LeftAbove);
      freq[imageData[j*(row_width+1)+1+3*i+component]]++; //  
      freq1[(UBYTE1)(imageData[j*(row_width+1)+1+3*i+component]-predict)]++; //   ϳ
     }
  lenEntropiN=lenEntropiCode(freq, 256);
  lenEntropi1=lenEntropiCode(freq1, 256);
  if (lenEntropi1<lenEntropiN) //  ϳ  
   {lenEntropi+=lenEntropi1;
    //     
    imageData[0]=4;
    imageDataPredict[0]=4;
    for (j=0; j<image_row; j+=step);
    for (j-=step; j>=0; j-=step)
     {for (i=0; i<image_col; i+=step);
      for (i-=step; i>=0; i-=step)
       for (component=0; component<=2; component++)
        {if (j>=step)
          Above=imageData[(j-step)*(row_width+1)+1+3*i+component];
         else
          Above=0;
         if (j>=step && i>=step)
          LeftAbove=imageData[(j-step)*(row_width+1)+1+3*(i-step)+component];
         else
          LeftAbove=0;
         if (i>=step)
          Left=imageData[j*(row_width+1)+1+3*(i-step)+component];
         else
          Left=0;
         predict=PaethPredictor(Left, Above, LeftAbove);
         imageDataPredict[j*(row_width+1)+1+3*i+component]=imageData[j*(row_width+1)+1+3*i+component]-predict;
        }
     }
   }
  else
   lenEntropi+=lenEntropiN;
  delete [] deltaPredict1; deltaPredict1=NULL;
  delete [] deltaPredict2; deltaPredict2=NULL;
  delete [] deltaPredict4; deltaPredict4=NULL;
  delete [] deltaPredict5; deltaPredict5=NULL;
  delete [] pprRow; pprRow=NULL;
  //    䳿 
  memset(freq,0,256*sizeof(long));
  for (j=0; j<image_row;j++)
   for (i=0; i<image_col;i++)
    {freq[imageDataPredict[j*(row_width+1)+1+3*i]]++;
     freq[imageDataPredict[j*(row_width+1)+1+3*i+1]]++;
     freq[imageDataPredict[j*(row_width+1)+1+3*i+2]]++;
    }
  entropi=entropiCode(freq, 256);
  GetLocalTime(&endTimeAnaliz);
  long int intervalAnaliz= (endTimeAnaliz.wHour-startTimeAnaliz.wHour)*60*60*1000 +
                     (endTimeAnaliz.wMinute-startTimeAnaliz.wMinute)*60*1000 +
                     (endTimeAnaliz.wSecond-startTimeAnaliz.wSecond)*1000 +
                     (endTimeAnaliz.wMilliseconds-startTimeAnaliz.wMilliseconds);
  cout << endl << "vubir predict " << intervalAnaliz << " milliseconds " ;
  cout << endl << "entropi " << entropi << " bpb" << endl ;
  cout << "entropi progres " << lenEntropi/(image_row*row_width) << " bpb" << endl ;
  //       
  UBYTE1 * imageDataP=new UBYTE1[image_row*(row_width+1)+500];
  UBYTE2 countBlockFilter=1+3*k;
  indexBlockFilter=0;
  nextPozStartBlockFilter = new UBYTE4[countBlockFilter+1];
  countPixelLineBlockFilter = new UBYTE2[countBlockFilter+1];
  stepBlockFilter = new UBYTE2[countBlockFilter+1]; //    
  proxidBlockFilter = new UBYTE1[countBlockFilter+1]; //      

  //  
  index=0;
  stepShar1=step;
  //  
  for (j=0; j<image_row; j++)
   imageDataP[index++]=imageData[j*(row_width+1)];
  //   ;
  for (j=0; j<image_row; j+=step)
   for (i=0; i<image_col; i+=step)
    for (component=0; component<=2; component++)
     imageDataP[index++]=imageData[j*(row_width+1)+1+3*i+component];
  countPixelLineBlockFilter[indexBlockFilter]=1+(image_col-1)/step;
  stepBlockFilter[indexBlockFilter]=step;
  proxidBlockFilter[indexBlockFilter]=1;
  nextPozStartBlockFilter[indexBlockFilter++]=index;
  //   
  for (l=2; l<=k+1; l++) //  
   {//    
    for (j=step/2; j<image_row; j+=step)
     for (i=step/2; i<image_col; i+=step)
      for (component=0; component<=2; component++)
       imageDataP[index++]=imageData[j*(row_width+1)+1+3*i+component];
    countPixelLineBlockFilter[indexBlockFilter]=1+(image_col-1-step/2)/step;
    stepBlockFilter[indexBlockFilter]=step;
    proxidBlockFilter[indexBlockFilter]=1;
    nextPozStartBlockFilter[indexBlockFilter++]=index;
    //     
    for (j=step/2; j<image_row; j+=step)
     for (i=0; i<image_col; i+=step)
      for (component=0; component<=2; component++)
       imageDataP[index++]=imageData[j*(row_width+1)+1+3*i+component];
    countPixelLineBlockFilter[indexBlockFilter]=1+(image_col-1)/step;
    stepBlockFilter[indexBlockFilter]=step;
    proxidBlockFilter[indexBlockFilter]=2;
    nextPozStartBlockFilter[indexBlockFilter++]=index;
    //     
    for (j=0; j<image_row; j+=step)
     for (i=step/2; i<image_col; i+=step)
      for (component=0; component<=2; component++)
       imageDataP[index++]=imageData[j*(row_width+1)+1+3*i+component];
    countPixelLineBlockFilter[indexBlockFilter]=1+(image_col-1-step/2)/step;
    stepBlockFilter[indexBlockFilter]=step;
    proxidBlockFilter[indexBlockFilter]=3;
    nextPozStartBlockFilter[indexBlockFilter++]=index;
    step/=2;
   }
  memcpy(imageData, imageDataP, index);

  //   䳿 
  index=0;
  step=stepShar1;
  //  
  for (j=0; j<image_row; j++)
   imageDataP[index++]=imageDataPredict[j*(row_width+1)];
  //   ;
  for (j=0; j<image_row; j+=step)
   for (i=0; i<image_col; i+=step)
    for (component=0; component<=2; component++)
     imageDataP[index++]=imageDataPredict[j*(row_width+1)+1+3*i+component];
  //   
  for (l=2; l<=k+1; l++) //  
   {//    
    for (j=step/2; j<image_row; j+=step)
     for (i=step/2; i<image_col; i+=step)
      for (component=0; component<=2; component++)
       imageDataP[index++]=imageDataPredict[j*(row_width+1)+1+3*i+component];
    //     
    for (j=step/2; j<image_row; j+=step)
     for (i=0; i<image_col; i+=step)
      for (component=0; component<=2; component++)
       imageDataP[index++]=imageDataPredict[j*(row_width+1)+1+3*i+component];
    //     
    for (j=0; j<image_row; j+=step)
     for (i=step/2; i<image_col; i+=step)
      for (component=0; component<=2; component++)
       imageDataP[index++]=imageDataPredict[j*(row_width+1)+1+3*i+component];
    step/=2;
   }
  memcpy(imageDataPredict, imageDataP, index);

  delete [] imageDataP; imageDataP=NULL;
  indexBlockFilter=0;
  countZamina=0;
  //        
  currentPozImage=0;
  // Fill in the image header.
  PngImageHeader header ;
  header.width = SystemToBigEndian ((UBYTE4) current_image->Width ()) ;
  header.height = SystemToBigEndian ((UBYTE4) current_image->Height ()) ;
  header.bitdepth = current_image->BitCount () ;
  if (current_image->BitCount () != 24)
  {
    // Determine if the image is grayscale.
    bool isgrayscale = true ;
    for (unsigned int ii = 0 ; ii < current_image->ColorCount () ; ++ ii)
    {
      if (current_image->ColorMap (ii).red != current_image->ColorMap (ii).green
          ||
          current_image->ColorMap (ii).red != current_image->ColorMap (ii).blue)
      {
        isgrayscale = false ;
        break ;
      }
    }

    if (isgrayscale)
      header.colortype = Grayscale ;
    else
      header.colortype = Palette ;
  }
  else
  {
    header.bitdepth = 8 ;
    header.colortype = RGB ;
  }
  header.compressionmethod = 0 ;
  header.filtermethod = 0 ;
  header.interlacemethod = 0 ;

  // Output the PNG Signature and header.
  output_stream->write ((char *) HbfSignature, HbfSignatureSize) ;
  StartChunk ("IHDR") ;
  OutputBlock (&header, sizeof (header)) ;
  WriteCurrentChunk () ;

  // Output any text blocks with keywords defined.
  //WriteTextBlocks () ;
  writeCM();
  // Output the palette if it is required.
  if (header.colortype == Palette)
  {
    StartChunk ("PLTE") ;
    for (unsigned int ii = 0 ; ii < current_image->ColorCount () ; ++ ii)
    {
      OutputByte (current_image->ColorMap (ii).red) ;
      OutputByte (current_image->ColorMap (ii).green) ;
      OutputByte (current_image->ColorMap (ii).blue) ;
    }
   WriteCurrentChunk () ;
  }

  WriteImageData () ;

  // Write the IEND marker.
  StartChunk ("IEND") ;
  WriteCurrentChunk () ;

  return ;
}

//
//  Description:
//
//    This function filters an image row.
//
//  Parameters:
//    row: The row to filter (1..imageheight - 1)
//

void PngEncoder::FilterRow (unsigned int row)
{
  // We need this because MSVC++ does not follow standard
  // scoping rules in for statements.
  unsigned int ii, filter1, filter2;
  if (filter_mask == 0)
    throw EPngError ("INTERNAL ERROR - No Filters Selected") ;

  //unsigned int mask = (1 << FilterNone) ;

  // Filter None
  for (ii = 0 ; ii < row_width ; ++ ii)
  {
    filter_buffers [FilterNone][ii] = (*current_image) [row][ii] ;
  }
  // If we only performed FilterNone then we do not need to proceed
  // any further.
  current_filter = FilterNone ;
  /*if (filter_mask == (1 << FilterNone))
   return ;
  // Filter for each type in the filter_mask.
  if ((filter_mask & (1 << FilterSub)) != 0)
  {
    mask |= (1 << FilterSub) ;

    UBYTE1 last ;
    for (unsigned int ii = 0 ; ii < row_width ; ++ ii)
    {
      if (ii >= countComponent)
        last = (*current_image)[row][ii-countComponent] ;
      else
        last = 0 ;

      filter_buffers [FilterSub][ii] = (*current_image) [row][ii] - last ;
    }
  }
  /*
  if ((filter_mask & (1 << FilterUp)) != 0 )
  {
    mask |= (1 << FilterUp) ;

    for (unsigned int ii = 0 ; ii < row_width ; ++ ii)
    {
      UBYTE1 above ;
      if (row == 0)
        above = 0 ;
      else
        above = (*current_image)[row-1][ii] ;
      filter_buffers [FilterUp][ii] = (*current_image) [row][ii] - above ;
    }
  }*/

/*  if ((filter_mask & (1 << FilterAverage)) != 0 && row > 0)
  {
    mask |= (1 << FilterAverage) ;

    UBYTE1 last ;
    UBYTE1 above ;
    for (unsigned int ii = 0 ; ii < row_width ; ++ ii)
    {
      if (ii >= countComponent)
        last = (*current_image)[row][ii - countComponent] ;
      else
        last = 0 ;
      if (row == 0)
        above = 0 ;
      else
        above = (*current_image) [row-1][ii] ;

      filter_buffers [FilterAverage][ii] =
        (*current_image) [row][ii] - (above + last) / 2 ;
    }
  }

  if ((filter_mask & (1 << FilterPaeth)) != 0 && row > 0)
  {
    mask |= (1 << FilterPaeth) ;

    UBYTE1 last ;
    UBYTE1 lastabove ;
    UBYTE1 above ;
    for (unsigned int ii = 0 ; ii < row_width ; ++ ii)
    {
      if (ii >= countComponent)
      {
        last = (*current_image)[row][ii-countComponent] ;
        if (row != 0)
          lastabove = (*current_image)[row-1][ii - countComponent] ;
        else
          lastabove = 0 ;
      }
      else
      {
        last = 0 ;
        lastabove = 0 ;
      }
      if (row == 0)
        above = 0 ;
      else
        above = (*current_image)[row-1][ii] ;
      filter_buffers [FilterPaeth][ii] = (*current_image) [row][ii]
                   - PaethPredictor (last, above, lastabove) ;
    }
  }*/
  current_filter=0;
  return;
}

//
//  Description:
//
//    This function specifies whether or not filters are used.
//
//  Parameters:
//    value: true=>Use Filters, false=>Don't use filters
//
void PngEncoder::SetUseFilters (bool value)
{
  if (value)
    filter_mask = 0xFFFFFFFF ;
  else
    filter_mask = 0x1 ;
}

//
//  Description:
//
//    This function initializes the Hash Table.
//
//
void PngEncoder::InitializeHashTable ()
{
  UBYTE4 ii ;
  memset (hash_values, 0, sizeof (HashEntry) * PngWindowSize) ;
  memset (hash_table, 0, sizeof (PHashEntry) * HashTableSize) ;
  memset (countHashValue, 0, sizeof (unsigned long) * HashTableSize) ;
  return ;
}

//
//  Description:
//
//    This function writes a Deflate block header and Huffman table
//    descriptions to the output stream.
//
//  Parameters:
//    lastblock: true => This is the last block in the image
//               false => There are more blocks to come
//
void PngEncoder::OutputDeflateHeader (bool lastblock)
{
  // Determine the count of length/literal and distances that
  // are used.
  unsigned int lengthcount ;
  for (lengthcount = PngMaxLengthCodes ;
       lengthcount > 0 ;
       -- lengthcount)
  {
    if (length_table->ehufsi [lengthcount - 1] != 0)
      break ;
  }

  unsigned int distancecount ;
  for (distancecount = PngMaxLengthCodes ;
       distancecount > 1 ;
       -- distancecount)
  {
    if (distance_table->ehufsi [distancecount - 1] != 0)
      break ;
  }

  // Gather the Huffman statistics for encoding the
  // lengths then create the Huffman table for doing so.
  length_length_table->Reset () ;
  FindLengthCodes (&PngEncoder::GatherLengthCounts,
                   length_table->ehufsi,
                   lengthcount) ;
  FindLengthCodes (&PngEncoder::GatherLengthCounts,
                   distance_table->ehufsi,
                   distancecount) ;
  length_length_table->BuildTable (PngMaxLengthLengthCodeSize) ;

  // Count the number of lengths we have to output.
  unsigned int hclen ;
  for (hclen = 19 ; hclen > 0 ; -- hclen)
  {
    if (length_length_table->ehufsi [PngLengthOrder [hclen-1]] != 0)
      break ;
  }

  // Write the Deflate header to the IDAT bock.
  if (lastblock)
    OutputDataBits (1, 1) ;
  else
    OutputDataBits (0, 1) ;

  OutputDataBits (2, 2) ; // Dynamic Huffman Codes
  OutputDataBits (lengthcount - 257, 5) ;
  OutputDataBits (distancecount - 1, 5) ;
  OutputDataBits (hclen - 4, 4) ;
  // Output the data for the Huffman table that encodes the Huffman tables
  for (unsigned int ii = 0 ; ii < hclen ; ++ ii)
    OutputDataBits (length_length_table->ehufsi [PngLengthOrder [ii]], 3) ;

   // Huffman encode the lengths for the Length/Literal and Distance
  // Huffman tables.
  FindLengthCodes (&PngEncoder::OutputLengthCounts,
                   length_table->ehufsi,
                   lengthcount) ;
  FindLengthCodes (&PngEncoder::OutputLengthCounts,
                   distance_table->ehufsi,
                   distancecount) ;
  return ;
}

//
//  Description:
//
//    This function writes a ZLIB header to the output stream.
//
void PngEncoder::OutputZLibHeader ()
{
  UBYTE1 cmf = 0x78 ;  // 7=>32K Sliding Window, 8=> Deflate Compression
  UBYTE1 flg = compression_level << 6 ;
  UBYTE2 check = (cmf << 8) | flg ;
  flg |= 31 - (check % 31) ;
  OutputDataBits (cmf, 8) ;
  OutputDataBits (flg, 8) ;
  return ;
}

//
//  Description:
//
//    This function writes the image data to a series of IDAT blocks.
//

void pidraxLenEntropiCode(UBYTE4 *masFreq, double *lenEntropiCode, unsigned int countAllFreq=256)
 {memset(lenEntropiCode, 0, countAllFreq*sizeof(double));
  UBYTE4 count=0, i;
  for (i=0; i<countAllFreq; i++)
   count+=masFreq[i];
  for (i=0; i<countAllFreq; i++)
   if (masFreq[i])
    lenEntropiCode[i]=log2((double)count/masFreq[i]);
   else
    lenEntropiCode[i]=16; // 
 }

void PngEncoder::WriteImageData ()
{
  // Write the IDAT blocks.
  StartChunk ("IDAT") ;

  adler_value = 1 ;

  block_buffer = new UBYTE2 [block_buffer_size] ;
  trebaZaminaBuffer = new UBYTE1 [block_buffer_size] ;
  if (trebaOptimizeLZBlock)
   {pozAnalizZamina = new UBYTE4 [block_buffer_size] ;
    pozImageBuffer = new UBYTE4 [block_buffer_size] ;
   }
  else
   memset(trebaZaminaBuffer, 1, block_buffer_size*sizeof(UBYTE1));
  currentPozImage=0;
  countByteImage=(row_width+1) * current_image->Height ();

  OutputZLibHeader () ;
  CallProgressFunction (0) ;
  // What we are doing here is outputing the image into a series of
  // Deflate compressed blocks.
  InitializeHashTable (); //    ProcessImageData
  for (bool nomoredata = false ; ! nomoredata ;)
  {
    // Save the input state
    length_table->StartEncode() ;
    plusBitBuffer->StartOutput();
    distance_table->StartEncode() ;
    countAnalizZamina=0;
    nomoredata = ProcessImageData () ;
    //  
    unsigned int freqCodeLengthTable[PngMaxLengthCodes], freqDistanceLengthTable[PngMaxDistanceCodes];
    if (trebaOptimizeLZBlock)
     {memcpy(freqCodeLengthTable, length_table->freq, PngMaxLengthCodes*sizeof(unsigned int));
      memcpy(freqDistanceLengthTable, distance_table->freq, PngMaxDistanceCodes*sizeof(unsigned int));
      unsigned int len, offset, i, j, k, countBitZamina;
      unsigned int code, extra , value, codeD, extraD , valueD;
      UBYTE4 poz;
      int buloZamina, nayavnoZamina;
      UBYTE1 lenTrebaZamina=minLenZaminaLiteral, pprNayavno=0;
      bool nayavnoLenZamina[minLenZaminaLiteral+1]; //      
      UBYTE4 freqProbaLength[minLenZaminaLiteral+1][286];
      UBYTE4 freqProbaDistance[minLenZaminaLiteral+1][30];
      UBYTE4 plusBit[minLenZaminaLiteral+1];
      memset(nayavnoLenZamina, 0, sizeof(nayavnoLenZamina));
      memset(freqProbaLength, 0, sizeof(freqProbaLength));
      memset(freqProbaDistance, 0, sizeof(freqProbaDistance));
      memset(plusBit, 0, sizeof(plusBit));
      //     
      for (i=0; i<countAnalizZamina; i++)
       {len = block_buffer[pozAnalizZamina[i]] - 256;
        offset=block_buffer[pozAnalizZamina[i]+1];
        LengthToCode (len, code, extra, value);
        DistanceToCode (offset, codeD, extraD, valueD) ;
        poz=pozImageBuffer[pozAnalizZamina[i]];
        if (len<minLenZaminaLiteral) //   
         {nayavnoLenZamina[len]=true;
          //     
          for (j=0; j<len; j++)
           freqProbaLength[len][imageDataPredict[poz+j]]++;
          freqProbaLength[len][code]++;
          freqProbaDistance[len][codeD]++;
          plusBit[len]+=extra+extraD;
         }
        else
         {nayavnoLenZamina[minLenZaminaLiteral]=true;
          //     
          for (j=0; j<len; j++)
           freqProbaLength[minLenZaminaLiteral][imageDataPredict[poz+j]]++;
          freqProbaLength[minLenZaminaLiteral][code]++;
          freqProbaDistance[minLenZaminaLiteral][codeD]++;
          plusBit[minLenZaminaLiteral]+=extra+extraD;
         }
       }
      double minSizeCode=0xFFFFFFFF, sizeCode, sizeLiteral, sizeZamina;
      UBYTE4 newCountAnalizZamina, freq;
      nayavnoLenZamina[2]=true; //      
      double lenProbaLength[286];
      double lenProbaDistance[30];
      UBYTE4 freqLL[286], freqD[30];
      for (k=2; k<minLenZaminaLiteral+1; k++) //      
       if (nayavnoLenZamina[k])
        {for (i=0; i<=256; i++)
          {freqProbaLength[k][i]+=freqProbaLength[pprNayavno][i]; //  ,  
           freqLL[i]=freqCodeLengthTable[i]+freqProbaLength[k][i];
          }
         for (i=257; i<=285; i++)
          {freqProbaLength[k][i]+=freqProbaLength[pprNayavno][i]; //  ,  
           freqLL[i]=freqCodeLengthTable[i]-freqProbaLength[k][i];
          }
         sizeCode=sizeEntropiCode(freqLL, 286);
         for (i=0; i<=29; i++)
          {freqProbaDistance[k][i]+=freqProbaDistance[pprNayavno][i]; //  ,  
           freqD[i]=freqDistanceLengthTable[i]-freqProbaDistance[k][i];
          }
         sizeCode+=sizeEntropiCode(freqD, 30);
         plusBit[k]+=plusBit[pprNayavno]; //  ,  
         sizeCode-=plusBit[k];
         if (sizeCode<minSizeCode)
          {minSizeCode=sizeCode;
           lenTrebaZamina=k;
          }
         pprNayavno=k;
        }
      //    
      for (i=0; i<=256; i++)
       freqLL[i]=freqCodeLengthTable[i]+freqProbaLength[lenTrebaZamina][i];
      for (i=257; i<=285; i++)
       freqLL[i]=freqCodeLengthTable[i]-freqProbaLength[lenTrebaZamina][i];
      pidraxLenEntropiCode(freqLL, lenProbaLength, 286);
      for (i=0; i<=29; i++)
       freqD[i]=freqDistanceLengthTable[i]-freqProbaDistance[lenTrebaZamina][i];
      pidraxLenEntropiCode(freqD, lenProbaDistance, 30);
      //     
      buloZamina=0;
      for (i=0; i<countAnalizZamina; i++)
       {len=block_buffer[pozAnalizZamina[i]] - 256;
        offset=block_buffer[pozAnalizZamina[i]+1];
        LengthToCode (len, code, extra, value);
        DistanceToCode (offset, codeD, extraD, valueD) ;
        poz=pozImageBuffer[pozAnalizZamina[i]];
        sizeZamina=lenProbaLength[code]+extra+lenProbaDistance[codeD]+extraD;
        sizeLiteral=0;
        for (j=0; j<len && sizeLiteral<sizeZamina; j++)
         sizeLiteral+=lenProbaLength[imageDataPredict[poz+j]];
        if (sizeLiteral<sizeZamina)
         {// 
          trebaZaminaBuffer[pozAnalizZamina[i]]=0;
          for (j=0; j<len; j++)
           freqCodeLengthTable[imageDataPredict[poz+j]]++;
          freqCodeLengthTable[code]--;
          freqDistanceLengthTable[codeD]--;
         }
        else
         buloZamina++;
       }
      //   
      int countIter=0;
      if (lenTrebaZamina>2)
       while (true)
        {//    
         for (i=0; i<=285; i++)
          freqLL[i]=freqCodeLengthTable[i];
         pidraxLenEntropiCode(freqLL, lenProbaLength, 286);
         for (i=0; i<=29; i++)
          freqD[i]=freqDistanceLengthTable[i];
         pidraxLenEntropiCode(freqD, lenProbaDistance, 30);
         //     
         nayavnoZamina=0;
         for (i=0; i<countAnalizZamina; i++)
          {len = block_buffer[pozAnalizZamina[i]] - 256;
           offset=block_buffer[pozAnalizZamina[i]+1];
           LengthToCode (len, code, extra, value);
           DistanceToCode (offset, codeD, extraD, valueD) ;
           poz=pozImageBuffer[pozAnalizZamina[i]];
           sizeZamina=lenProbaLength[code]+extra+lenProbaDistance[codeD]+extraD;
           sizeLiteral=0;
           if (trebaZaminaBuffer[pozAnalizZamina[i]]) // 
            {for (j=0; j<len && sizeLiteral<sizeZamina; j++)
              sizeLiteral+=lenProbaLength[imageDataPredict[poz+j]];
             if (sizeLiteral<sizeZamina)
              {// 
               trebaZaminaBuffer[pozAnalizZamina[i]]=0;
               for (j=0; j<len; j++)
                freqCodeLengthTable[imageDataPredict[poz+j]]++;
               freqCodeLengthTable[code]--;
               freqDistanceLengthTable[codeD]--;
              }
            }
           else // 
            {for (j=0; j<len; j++)
              sizeLiteral+=lenProbaLength[imageDataPredict[poz+j]];
             if (sizeLiteral>sizeZamina)
              {// 
               trebaZaminaBuffer[pozAnalizZamina[i]]=1;
               for (j=0; j<len; j++)
                freqCodeLengthTable[imageDataPredict[poz+j]]--;
               freqCodeLengthTable[code]++;
               freqDistanceLengthTable[codeD]++;
              }
            }
           if (trebaZaminaBuffer[pozAnalizZamina[i]])
            nayavnoZamina++;
          }
         if (++countIter>1 && buloZamina-nayavnoZamina<10)
          break; //    
         buloZamina=nayavnoZamina;
        }
      memcpy(length_table->freq, freqCodeLengthTable,  PngMaxLengthCodes*sizeof(unsigned int));
      memcpy(distance_table->freq, freqDistanceLengthTable, PngMaxDistanceCodes*sizeof(unsigned int));
     }
    if (trebaStatisticAnalizeLZBlock)
     SAOutputBlockData (nomoredata) ;
    else
     {length_table->BuildTable () ;
      distance_table->BuildTable () ;
      OutputDeflateHeader (nomoredata) ;
      OutputBlockData ();
     }
    CallProgressFunction (currentPozImage*100/countByteImage);
  }

  CallProgressFunction (100) ;
  OutputAdler () ;
  WriteCurrentChunk () ;

  delete [] block_buffer ; block_buffer = NULL ;
  delete [] trebaZaminaBuffer ; trebaZaminaBuffer = NULL ;
  if (trebaOptimizeLZBlock)
   {delete [] pozAnalizZamina; pozAnalizZamina = NULL;
    delete [] pozImageBuffer; pozImageBuffer = NULL;
   }
  return ;
}

void PngEncoder::OutputDataByte(UBYTE1 *buffer, UBYTE2 length)
 {if (!length)
   return;
  if (bit_position != 0)
   {++ byte_position;
    bit_position=0;
    if (byte_position >= OutputBufferSize)
     {WriteCurrentChunk () ;
      StartChunk ("IDAT") ;
     }
   }
  if (byte_position+length<OutputBufferSize) //   
   {memcpy(output_buffer+byte_position, buffer, length);
    byte_position+=length;
   }
  else
   {UBYTE2 poz=OutputBufferSize-byte_position;
    memcpy(output_buffer+byte_position, buffer, poz);
    byte_position=OutputBufferSize;
    WriteCurrentChunk () ;
    StartChunk("IDAT") ;
    while (poz+OutputBufferSize<=length)
     {memcpy(output_buffer, buffer+poz, OutputBufferSize);
      byte_position=OutputBufferSize;
      WriteCurrentChunk () ;
      StartChunk("IDAT") ;
      poz+=OutputBufferSize;
     }
    memcpy(output_buffer, buffer+poz, length-poz);
    byte_position=length-poz;
   }
  output_buffer [byte_position] = 0;
  return ;
}

//
//  Description:
//
//    This function Huffman encodes and outputs the buffered data.
//
//    The buffer is encoded so that
//
//      0..255  is a literal byte
//      256-514 is a length code of N-256
//
//    Each length code is followed by a distance code.
//
void PngEncoder::OutputBlockData ()
{
  UBYTE2 huffmancode ;
  UBYTE1 huffmansize ;
  unsigned int code ;
  unsigned int extra ;
  unsigned int value ;
  unsigned int limit = block_buffer_count ;
  for (unsigned int ii = 0 ; ii < limit ; ++ ii)
  {
    if (block_buffer [ii] < 256)
    {
      length_table->Encode (block_buffer [ii]) ;
    }
    else
    {
      unsigned int length = block_buffer [ii] - 256 ;
      if (trebaZaminaBuffer[ii])
       {LengthToCode (length, code, extra, value) ;
        length_table->Encode (code) ;
        if (extra != 0)
         plusBitBuffer->OutputBit(value, extra) ;

        ++ ii ;
        unsigned int distance = block_buffer [ii] ;
        DistanceToCode (distance, code, extra, value) ;
        distance_table->Encode (code) ;
        if (extra != 0)
         plusBitBuffer->OutputBit(value, extra) ;
       }
      else
       {UBYTE4 poz=pozImageBuffer[ii];
        for (unsigned int j = 0; j<length; j++)
         length_table->Encode (imageDataPredict[poz+j]) ;
        ++ii;
       }
    }
  }
  length_table->Encode (PngEndCode) ;
  length_table->FinishEncode();
  distance_table->FinishEncode();
  //   
  OutputDataByte(length_table->bufFreq, length_table->pozBufFreq);
  OutputDataBits(length_table->pozBuffer, 16);
  OutputDataByte(length_table->buffer, length_table->pozBuffer);
  if (distance_table->nayavnoCode)
   {OutputDataByte(distance_table->bufFreq, distance_table->pozBufFreq);
    OutputDataBits(distance_table->pozBuffer, 16);
    OutputDataByte(distance_table->buffer, distance_table->pozBuffer);
   }
  OutputDataByte(plusBitBuffer->buffer, plusBitBuffer->pozBuffer);
  OutputDataBits(plusBitBuffer->buffer[plusBitBuffer->pozBuffer], plusBitBuffer->pozBitBuffer);
  return ;
}


UBYTE2 sabsMinus(UBYTE2 a, UBYTE2 b)
 {if (a>b)
   return a-b;
  else
   return b-a;
 }

UBYTE4 lmin(UBYTE4 a, UBYTE4 b)
 {if (a>b)
   return b;
  else
   return a;
 }

UBYTE4 lmax(UBYTE4 a, UBYTE4 b)
 {if (a>b)
   return a;
  else
   return b;
 }

typedef unsigned int freqLL[PngMaxLengthCodes];
typedef unsigned int freqD[PngMaxDistanceCodes];

UBYTE4 countBitCodeFreqLL(unsigned int * masFreq, UBYTE4 freqElementMas)
 {double s=0;
  UBYTE2 countElement=0;
  for (UBYTE2 i=0; i<PngMaxLengthCodes; i++)
   if (masFreq[i]>0)
    {countElement++;
     if (masFreq[i]>1)
      s+=masFreq[i]*log2(masFreq[i]);
    }
  return 70+countElement*log2(countElement)+freqElementMas*log2(freqElementMas)-s;
 }

UBYTE4 countBitCodeSumisnFreqLL(freqLL &masFreq1, freqLL &masFreq2, UBYTE4 freqElementMas)
 {double s=0;
  UBYTE2 countElement=0;
  for (UBYTE2 i=0; i<PngMaxLengthCodes; i++)
   if (masFreq1[i]+masFreq2[i]>0)
    {countElement++;
     if (masFreq1[i]+masFreq2[i]>1)
      s+=(masFreq1[i]+masFreq2[i])*log2(masFreq1[i]+masFreq2[i]);
    }
  return 70+countElement*log2(countElement)+freqElementMas*log2(freqElementMas)-s;
 }

void PngEncoder::SAOutputBlockData (bool lastblock)
 {UBYTE4 countElement=0, i, j, k, poz;
  UBYTE4 countABlock, startPozABlock, endPozABlock; //  
  UBYTE4 countRBlock, startPozRBlock, endPozRBlock; //  
  unsigned int code, extra, value;
  UBYTE2 length, cf, maxFBlock, maxFBlock1;
  UBYTE4 indexMaxFBlock, indexMaxFBlock1;
  for (i=0; i<286; i++)
   countElement+=length_table->freq[i];
  countElement-=1; //    
  if (countElement<1000) // 
   {length_table->BuildTable () ;
    distance_table->BuildTable () ;
    OutputDeflateHeader(lastblock) ;
    OutputBlockData();
    return;
   }
  UBYTE2 * codeLL = new UBYTE2[countElement];
  UBYTE2 * codeD = new UBYTE2[countElement];
  //  
  i=0;
  for (j = 0 ; j < block_buffer_count ; j++)
   {if (block_buffer[j] < 256)
     codeLL[i++]=block_buffer[j];
    else
     if (trebaZaminaBuffer[j])
      {codeLL[i]=block_buffer[j++];
       codeD[i++]=block_buffer[j];
      }
     else
      {length = block_buffer[j] - 256 ;
       poz=pozImageBuffer[j];
       for (k = 0; k<length; k++)
        codeLL[i++]=imageDataPredict[poz+k];
       j++;//  
      }
   }
  const lenWindow=100, maxLenABlock=0x4000;
  UBYTE2 huffmancode;
  UBYTE1 huffmansize;
  countABlock=(countElement-2*lenWindow+maxLenABlock-1)/maxLenABlock;
  UBYTE1 countLeft[PngMaxLengthCodes], countRight[PngMaxLengthCodes];
  UBYTE1 f[maxLenABlock], indexMaxEkonom;
  UBYTE4 * startRBlock = new UBYTE4[countABlock*3+2];
  int ekonomBit;
  //  
  memset(countLeft, 0, PngMaxLengthCodes);
  memset(countRight, 0, PngMaxLengthCodes);
  for (j=0; j<lenWindow; j++)
   {if (codeLL[j]<256)
     code=codeLL[j];
    else
     LengthToCode (codeLL[j]-256, code, extra, value) ;
    countLeft[code]++;
    if (codeLL[j+lenWindow]<256)
     code=codeLL[j+lenWindow];
    else
     LengthToCode (codeLL[j+lenWindow]-256, code, extra, value) ;
    countRight[code]++;
   }
  cf=0;
  for (j=0; j<PngMaxLengthCodes; j++)
   cf+=sabsMinus(countLeft[j], countRight[j]);
  //       
  countRBlock=1;
  startRBlock[0]=0; //   
  for (i=0; i<countABlock; i++)
   {//      
    startPozABlock=lenWindow+i*maxLenABlock;
    endPozABlock=lmin(startPozABlock+maxLenABlock-1, countElement-1-lenWindow);
    maxFBlock=0;
    for (j=startPozABlock; j<=endPozABlock; j++)
     {//    
      if (codeLL[j-lenWindow]<256)
       code=codeLL[j-lenWindow];
      else
       LengthToCode (codeLL[j-lenWindow]-256, code, extra, value);
      cf-=sabsMinus(countLeft[code], countRight[code]);
      countLeft[code]--;
      cf+=sabsMinus(countLeft[code], countRight[code]);
      //    
      if (codeLL[j+lenWindow]<256)
       code=codeLL[j+lenWindow];
      else
       LengthToCode (codeLL[j+lenWindow]-256, code, extra, value);
      cf-=sabsMinus(countLeft[code], countRight[code]);
      countRight[code]++;
      cf+=sabsMinus(countLeft[code], countRight[code]);
      //      
      if (codeLL[j]<256)
       code=codeLL[j];
      else
       LengthToCode (codeLL[j]-256, code, extra, value);
      cf-=sabsMinus(countLeft[code], countRight[code]);
      countLeft[code]++;
      countRight[code]--;
      cf+=sabsMinus(countLeft[code], countRight[code]);
      //     
      if (cf>maxFBlock)
       {maxFBlock=cf;
        indexMaxFBlock=j;
       }
      f[j-startPozABlock]=cf;
     }
    if (maxFBlock) // 
     {//      
      if (indexMaxFBlock-lenWindow>=lmax(startPozABlock, startRBlock[countRBlock-1]+lenWindow))
       {maxFBlock1=0;
        for (j=lmax(startPozABlock, startRBlock[countRBlock-1]+lenWindow); j<=indexMaxFBlock-lenWindow; j++)
         if (f[j-startPozABlock]>maxFBlock1)
          {maxFBlock1=f[j-startPozABlock];
           indexMaxFBlock1=j;
          }
        //    
        if (maxFBlock1)
         startRBlock[countRBlock++]=indexMaxFBlock1+1;
       }
      //   
      startRBlock[countRBlock++]=indexMaxFBlock+1;
      //      
      if (indexMaxFBlock+lenWindow<=endPozABlock)
       {maxFBlock1=0;
        for (j=indexMaxFBlock+lenWindow; j<=endPozABlock; j++)
         if (f[j-startPozABlock]>maxFBlock1)
          {maxFBlock1=f[j-startPozABlock];
           indexMaxFBlock1=j;
          }
        //    
        if (maxFBlock1)
         startRBlock[countRBlock++]=indexMaxFBlock1+1;
       }
     }
   }
  startRBlock[countRBlock]=countElement; //   
  freqLL * freqLLRBlock=new freqLL[countRBlock];
  freqD * freqDRBlock=new freqD[countRBlock];
  UBYTE4 * countBitCodeRBlock=new UBYTE4[countRBlock];
  UBYTE4 * countBitCodeSRBlock=new UBYTE4[countRBlock];
  memset(freqLLRBlock, 0, countRBlock*sizeof(freqLL));
  memset(freqDRBlock, 0, countRBlock*sizeof(freqD));
  //        
  for (i=0; i<countRBlock; i++)
   {startPozRBlock=startRBlock[i];
    endPozRBlock=startRBlock[i+1]-1;
    for (j=startPozRBlock; j<=endPozRBlock; j++)
     if (codeLL[j]<256)
      freqLLRBlock[i][codeLL[j]]++;
     else
      {LengthToCode (codeLL[j]-256, code, extra, value);
       freqLLRBlock[i][code]++;
       DistanceToCode (codeD[j], code, extra, value);
       freqDRBlock[i][code]++;
      }
    countBitCodeRBlock[i]=countBitCodeFreqLL(freqLLRBlock[i], endPozRBlock-startPozRBlock+2);//    
   }
  //        
  for (i=0; i<countRBlock-1; i++)
   countBitCodeSRBlock[i]=countBitCodeSumisnFreqLL(freqLLRBlock[i], freqLLRBlock[i+1], startRBlock[i+2]-startRBlock[i]+1);//    
  //      
  while (countRBlock>1)
   {ekonomBit=countBitCodeRBlock[0]+countBitCodeRBlock[1]-countBitCodeSRBlock[0];
    indexMaxEkonom=0;
    for (i=1; i<countRBlock-1; i++)
     if ((int)(countBitCodeRBlock[i]+countBitCodeRBlock[i+1]-countBitCodeSRBlock[i])>ekonomBit)
      {ekonomBit=countBitCodeRBlock[i]+countBitCodeRBlock[i+1]-countBitCodeSRBlock[i];
       indexMaxEkonom=i;
      }
    if (ekonomBit<0)
     break;
    //    
    for (j=0; j<PngMaxLengthCodes; j++)
     freqLLRBlock[indexMaxEkonom][j]+=freqLLRBlock[indexMaxEkonom+1][j];
    memcpy(freqLLRBlock+indexMaxEkonom+1, freqLLRBlock+indexMaxEkonom+2, (countRBlock-indexMaxEkonom-2)*sizeof(freqLL));
    for (j=0; j<PngMaxDistanceCodes; j++)
     freqDRBlock[indexMaxEkonom][j]+=freqDRBlock[indexMaxEkonom+1][j];
    memcpy(freqDRBlock+indexMaxEkonom+1, freqDRBlock+indexMaxEkonom+2, (countRBlock-indexMaxEkonom-2)*sizeof(freqD));
    memcpy(startRBlock+indexMaxEkonom+1, startRBlock+indexMaxEkonom+2, (countRBlock-indexMaxEkonom-1)*sizeof(UBYTE4));
    countBitCodeRBlock[indexMaxEkonom]=countBitCodeFreqLL(freqLLRBlock[indexMaxEkonom], startRBlock[indexMaxEkonom+1]-startRBlock[indexMaxEkonom]+1);//    
    memcpy(countBitCodeRBlock+indexMaxEkonom+1, countBitCodeRBlock+indexMaxEkonom+2, (countRBlock-indexMaxEkonom-2)*sizeof(UBYTE4));
    if (indexMaxEkonom>0)
     countBitCodeSRBlock[indexMaxEkonom-1]=countBitCodeSumisnFreqLL(freqLLRBlock[indexMaxEkonom-1], freqLLRBlock[indexMaxEkonom], startRBlock[indexMaxEkonom+1]-startRBlock[indexMaxEkonom-1]+1);//    
    if (indexMaxEkonom+2<countRBlock)
     countBitCodeSRBlock[indexMaxEkonom]=countBitCodeSumisnFreqLL(freqLLRBlock[indexMaxEkonom], freqLLRBlock[indexMaxEkonom+1], startRBlock[indexMaxEkonom+2]-startRBlock[indexMaxEkonom]+1);//    
    if (indexMaxEkonom+3<countRBlock)
     memcpy(countBitCodeSRBlock+indexMaxEkonom+1, countBitCodeSRBlock+indexMaxEkonom+2, (countRBlock-indexMaxEkonom-3)*sizeof(UBYTE4));
    countRBlock--;
   }
  if (countRBlock>2) //     
   {k=0;
    for (i=0; i<countRBlock; i++)
     k+=countBitCodeRBlock[i];
    memset(length_table->freq, 0, 286*sizeof(unsigned int));
    for (i=0; i<countElement; i++)
     {if (codeLL[i]<256)
       code=codeLL[i];
      else
       LengthToCode (codeLL[i]-256, code, extra, value);
      length_table->freq[code]++;
     }
    if (k>=countBitCodeFreqLL(length_table->freq, countElement+1))//    
     {length_table->freq[256]=1;
      countRBlock=1;
     }
   }
  if (countRBlock>1) //   
   for (i=0; i<countRBlock; i++)
    {length_table->StartEncode() ;
     plusBitBuffer->StartOutput();
     distance_table->StartEncode() ;
     memcpy(length_table->freq, freqLLRBlock[i],  PngMaxLengthCodes*sizeof(unsigned int));
     length_table->freq[256]=1; //  
     memcpy(distance_table->freq, freqDRBlock[i], PngMaxDistanceCodes*sizeof(unsigned int));
     length_table->BuildTable () ;
     distance_table->BuildTable () ;
     if (i==countRBlock-1)
      OutputDeflateHeader (lastblock);
     else
      OutputDeflateHeader (false);
     startPozRBlock=startRBlock[i];
     endPozRBlock=startRBlock[i+1]-1;
     for (j=startPozRBlock; j<=endPozRBlock; j++)
      if (codeLL[j]<256)
       {length_table->Encode (codeLL[j]);
       }
      else
       {LengthToCode (codeLL[j]-256, code, extra, value);
        length_table->Encode (code);
        if (extra != 0)
         plusBitBuffer->OutputBit(value, extra) ;
        DistanceToCode (codeD[j], code, extra, value) ;
        distance_table->Encode (code) ;
        if (extra != 0)
         plusBitBuffer->OutputBit(value, extra) ;
       }
     length_table->Encode (PngEndCode);
     length_table->FinishEncode();
     distance_table->FinishEncode();
     //   
     OutputDataByte(length_table->bufFreq, length_table->pozBufFreq);
     OutputDataBits(length_table->pozBuffer, 16);
     OutputDataByte(length_table->buffer, length_table->pozBuffer);
     if (distance_table->nayavnoCode)
      {OutputDataByte(distance_table->bufFreq, distance_table->pozBufFreq);
       OutputDataBits(distance_table->pozBuffer, 16);
       OutputDataByte(distance_table->buffer, distance_table->pozBuffer);
      }
     OutputDataByte(plusBitBuffer->buffer, plusBitBuffer->pozBuffer);
     OutputDataBits(plusBitBuffer->buffer[plusBitBuffer->pozBuffer], plusBitBuffer->pozBitBuffer);
    }
  delete [] countBitCodeSRBlock; countBitCodeSRBlock=NULL;
  delete [] countBitCodeRBlock; countBitCodeRBlock=NULL;
  delete [] freqLLRBlock; freqLLRBlock=NULL;
  delete [] freqDRBlock; freqDRBlock=NULL;
  delete [] startRBlock; startRBlock=NULL;
  delete [] codeLL; codeLL=NULL;
  delete [] codeD; codeD=NULL;
  if (countRBlock==1) //  
   {length_table->BuildTable () ;
    distance_table->BuildTable () ;
    OutputDeflateHeader (lastblock) ;
    OutputBlockData();
   }
 }











